"""Contains all automatically generated Monoids from CFFI.

The scalar addition of conventional matrix multiplication is replaced
with a *monoid*.  A monoid is an associative and commutative binary
operator `z=f(x,y)` where all three domains are the same (the
types of `x`, `y`, and `z`), and where the operator has
an identity value `id` such that `f(x,id)=f(id,x)=x`.
Performing matrix multiplication with a semiring uses a monoid in
place of the `add` operator, scalar addition being just one of many
possible monoids.  The identity value of addition is zero, since
$x+0=0+x=x$.  GraphBLAS includes many built-in operators suitable for
use as a monoid: min (with an identity value of positive infinity),
max (whose identity is negative infinity), add (identity is zero),
multiply (with an identity of one), four logical operators: AND, OR,
exclusive-OR, and Boolean equality (XNOR), four bitwise operators
(AND, OR, XOR, and XNOR), and the ANY operator.  User-created monoids
can be defined with any associative and commutative operator that has
an identity value.
"""

__all__ = ["Monoid", "current_monoid"]

import os
import sys
import re
import contextvars
from itertools import chain
from collections import defaultdict

from .base import lib, ffi, _check
from .binaryop import BinaryOp
from . import types

current_monoid = contextvars.ContextVar("current_monoid")


class Monoid:

    __slots__ = ("name", "monoid", "token", "op", "type")

    def __init__(self, op, typ, monoid, udt=None, boolean=False):
        self.monoid = monoid
        cls = getattr(types, typ, None)
        if cls is not None:
            setattr(cls, op + "_MONOID", self)
            setattr(cls, op.lower() + "_monoid", self)
            types.__pdoc__[f"{typ}.{op}_MONOID"] = f"UnaryOp {typ}.{op}_MONOID"
        self.op = op
        self.type = typ
        self.name = "_".join((op, typ, "monoid"))
        self.token = None

    def __enter__(self):
        self.token = current_monoid.set(self)
        return self

    def __exit__(self, *errors):
        current_monoid.reset(self.token)
        return False

    def get_monoid(self, left=None, right=None):
        return self.monoid

    def print(self, level=2, name="", f=sys.stdout):  # pragma: nocover
        """Print the matrix using `GxB_Matrix_fprint()`, by default to
        `sys.stdout`.

        Level 1: Short description
        Level 2: Short list, short numbers
        Level 3: Long list, short number
        Level 4: Short list, long numbers
        Level 5: Long list, long numbers

        """
        _check(lib.GxB_Monoid_fprint(self.monoid, bytes(name, "utf8"), level, f))


gxb_monoid_re = re.compile(
    "^GxB_(MIN|MAX|PLUS|TIMES|ANY|BOR|BAND|BXOR|BXNOR)_"
    "(UINT8|UINT16|UINT32|UINT64|INT8|INT16|INT32|INT64|FP32|FP64|FC32|FC64)_MONOID$"
)

grb_monoid_re = re.compile(
    "^GrB_(MIN|MAX|PLUS|TIMES)_MONOID_"
    "(UINT8|UINT16|UINT32|UINT64|INT8|INT16|INT32|INT64|FP32|FP64|FC32|FC64)$"
)

pure_bool_re = re.compile("^GxB_(ANY|LOR|LAND|LXOR|LXNOR|EQ)_(BOOL)_MONOID$")
pure_bool_re_v13 = re.compile("^GrB_(LOR|LAND|LXOR|LXNOR)_MONOID_(BOOL)$")


def monoid_group(reg):
    srs = []
    for n in filter(None, [reg.match(i) for i in dir(lib)]):
        op, typ = n.groups()
        m = Monoid(op, typ, getattr(lib, n.string))
        srs.append(m)
    return srs


def build_monoids(__pdoc__):
    import tempfile

    this = sys.modules[__name__]
    for r in chain(
        monoid_group(gxb_monoid_re),
        monoid_group(grb_monoid_re),
        monoid_group(pure_bool_re),
        monoid_group(pure_bool_re_v13),
    ):
        setattr(this, r.name, r)
        f = tempfile.TemporaryFile()
        r.print(f=f)
        f.seek(0)
        this.__all__.append(r.name)
        op, typ, _ = r.name.split("_")
        __pdoc__[f"{typ}.{op}_MONOID"] = f"""```{str(f.read(), 'utf8')}```"""



"""Helper functions for drawing graphs and matrices with graphviz in
doctests and Jupyter notebooks.

>>> from pygraphblas import *
>>> M = Matrix.random(UINT8, 4, 4, 15, seed=8)

>>> g = draw_graph(M, show_weight=False,
...     filename='docs/imgs/Matrix_from_lists3')

![Matrix_from_lists3.png](../imgs/Matrix_from_lists3.png)

>>> g = draw_matrix(M, scale=50,
...     filename='docs/imgs/Matrix_from_lists4')

![Matrix_from_lists3.png](../imgs/Matrix_from_lists4.png)

>>> V = Vector.from_lists([0, 2], [3.14, 1.2])
>>> g = draw_vector(V, scale=50,
...     filename='docs/imgs/Vector_from_lists_1')

![Vector_from_lists_1](../imgs/Vector_from_lists_1.png)

>>> g = draw_matrix_op(M, '@', M, (M@M), scale=50,
...     filename='docs/imgs/mxm1')

![mxm1](../imgs/mxm1.png)
"""
import operator
from math import log
from itertools import accumulate
from graphviz import Graph, Digraph, Source
from PIL import Image, ImageDraw, ImageFont
from PIL import Image, ImageDraw
from json import dumps
from pathlib import Path
import matplotlib.pyplot as plt
from matplotlib.colors import rgb2hex
from collections import OrderedDict

__all__ = [
    "draw",
    "draw_graph",
    "draw_matrix",
    "draw_vector",
    "draw_layers",
    "draw_graph_op",
    "draw_matrix_op",
    "draw_matrix_layers",
    "draw_cy",
    "draw_vis",
]


def _str(s, label_width):  # pragma: nocover
    return str(s)[:label_width]


def draw_vector_dot(V, name="", rankdir="LR", ioff=0, joff=0):  # pragma: nocover
    g = Digraph(name)
    g.attr(rankdir=rankdir, ranksep="1")
    for i, v in V:
        g.node(str(i + ioff), label="%s:%s" % (str(i), str(v)))
    return g


def draw_graph(
    M,
    name="",
    rankdir="LR",
    directed=True,
    show_weight=True,
    concentrate=True,
    labels=True,
    label_vector=None,
    label_width=None,
    label_cmap=None,
    size_vector=None,
    size_scale=1.0,
    min_size=0.1,
    log_scale=False,
    ioff=0,
    joff=0,
    filename=None,
    graph_attr=None,
    node_attr=None,
    edge_attr=None,
    edge_cmap=None,
):  # pragma: nocover
    if directed:
        g = Digraph(name)
    else:
        g = Graph(name)
    g.attr(
        rankdir=rankdir, overlap="false", concentrate="true" if concentrate else "false"
    )
    if graph_attr:
        g.attr(**graph_attr)
    if node_attr:
        g.attr("node", **node_attr)
    if edge_attr:
        g.attr("edge", **edge_attr)

    if edge_cmap is not None:
        edge_cmap = plt.get_cmap(edge_cmap)

    if label_cmap is not None:
        label_cmap = plt.get_cmap(label_cmap)

    if isinstance(label_vector, list):
        labeler = lambda v, i: v[i] if labels else ""
    else:
        labeler = lambda v, i: v.get(i) if labels else ""

    for i, j, v in M:
        ilabel = (
            _str(labeler(label_vector, i), label_width)
            if label_vector
            else str(i)
            if labels
            else ""
        )
        jlabel = (
            _str(labeler(label_vector, j), label_width)
            if label_vector
            else str(j)
            if labels
            else ""
        )
        vlabel = _str(v, label_width) if show_weight else None

        if size_vector:
            scale = max(size_vector[i] * size_scale, min_size)
            if log_scale:
                scale = max(log(scale), min_size)
            size = _str(scale, label_width)
        else:
            size = "0.5"

        args = {"width": size, "fixedsize": "true"}
        if node_attr:
            args.update(node_attr)
        if labels:
            args["label"] = ilabel
        if label_cmap and label_vector:
            args["color"] = rgb2hex(label_cmap(float(labeler(label_vector, i))))
        inode = g.node(str(i + ioff), **args)

        args = {"width": size, "fixedsize": "true"}
        if node_attr:
            args.update(node_attr)
        if labels:
            args["label"] = jlabel
        if label_cmap:
            args["color"] = rgb2hex(label_cmap(float(labeler(label_vector, j))))
        jnode = g.node(str(j + joff), **args)
        w = str(v)

        args = {}
        if edge_attr:
            args.update(edge_attr)
        if labels:
            args["label"] = vlabel
            args["tooltip"] = vlabel
        if edge_cmap:
            args["color"] = rgb2hex(edge_cmap(float(v)))
        g.edge(str(i + ioff), str(j + joff), weight=w, **args)
    if filename is not None:
        g.render(filename, format="png")
    return g


def draw_layers(M, name="", rankdir="LR", label_width=None):  # pragma: nocover
    g = Digraph(name)
    g.attr(rankdir=rankdir, ranksep="1")
    for l, m in enumerate(M):
        with g.subgraph() as s:
            s.attr(rank="same", rankdir="TB")
            for i in range(m.nrows):
                si = (l * m.nrows) + i
                s.node(str(si), label=_str(si, label_width), width="0.5")
                if i < m.nrows - 1:
                    ni = si + 1
                    s.edge(str(si), str(ni), style="invis", minlen="0", weight="1000")
            g.edge(
                str(si - m.nrows + 1),
                str(si + 1),
                weight="10000000",
                style="invis",
                minlen="0",
            )

    with g.subgraph() as s:
        s.attr(rank="same", rankdir="LR")
        for j in range(M[-1].nrows):
            sj = (len(M) * m.nrows) + j
            s.node(str(sj), label=_str(j, label_width), width="0.5")
            if j < M[-1].nrows - 1:
                nj = sj + 1
                s.edge(str(sj), str(nj), style="invis")

    for l, m in enumerate(M):
        for i, j, _ in m:
            si = (l * m.nrows) + i
            sj = ((l + 1) * m.nrows) + j
            g.edge(str(si), str(sj))
    return g


def draw(obj, name="", **kws):  # pragma: nocover
    from pygraphblas import Matrix, Vector

    if isinstance(obj, Matrix):
        return draw_graph(obj, name, **kws)
    if isinstance(obj, Vector):
        return draw_vector_dot(obj, name, **kws)


def draw_graph_op(left, op, right, result, **kwargs):  # pragma: nocover
    from pygraphblas import Matrix, Vector

    ioff = 0
    joff = 0

    def draw(obj, name="", **kws):
        nonlocal ioff, joff
        if isinstance(obj, Matrix):
            ioff += obj.nrows
            joff += obj.ncols
            return draw_graph(obj, name=name, ioff=ioff, joff=joff)
        if isinstance(obj, Vector):
            ioff += obj.size
            joff += obj.size
            return draw_vector_dot(obj, name=name, ioff=ioff, joff=joff)

    g = Digraph()
    g.subgraph(draw(left, name="cluster_left"))
    g.node(op, width="0.5")
    g.subgraph(draw(right, name="cluster_right"))
    g.node("=", width="0.5")
    g.subgraph(draw(result, name="cluster_result"))
    return g


def draw_matrix(
    M,
    scale=10,
    axes=True,
    axes_labels=False,
    labels=True,
    mode=None,
    cmap="viridis",
    filename=None,
    column=True,
    font_path=None,
):  # pragma: nocover
    from pygraphblas import BOOL, FP32, FP64, Matrix, Vector

    font = None
    if font_path:
        font = ImageFont.truetype(
            str(font_path / "FantasqueSansMono-Bold.ttf"), int(scale * 0.5)
        )
    else:
        font = ImageFont.load_default()

    if isinstance(M, Vector):
        return draw_vector(
            M,
            scale=scale,
            axes=axes,
            axes_labels=axes_labels,
            labels=labels,
            mode=mode,
            cmap=cmap,
            filename=filename,
            column=column,
        )

    if mode is None:
        mode = "RGB"

    if cmap is not None:
        cmap = plt.get_cmap(cmap)

    sx = ((M.ncols + 1) * scale) + 1
    sy = ((M.nrows + 1) * scale) + 1
    im = Image.new(mode, (sx, sy), color="white")
    d = ImageDraw.Draw(im)
    for i, j, v in M:
        y = ((i + 1) * scale) + scale / 2
        x = ((j + 1) * scale) + scale / 2
        offset = int(scale / 2)
        if M.type is BOOL:
            d.rectangle(
                (x - offset, y - offset, x + scale - offset, y + scale - offset),
                fill=rgb2hex(cmap(1)) if v else rgb2hex(cmap(0)),
                outline="black",
            )
            v = "t" if v else "f"
        elif M.type in [FP32, FP64] and cmap:
            d.rectangle(
                (x - offset, y - offset, x + scale - offset, y + scale - offset),
                fill=rgb2hex(cmap(v)),
                outline="black",
            )
        else:
            d.rectangle(
                (x - offset, y - offset, x + scale - offset, y + scale - offset),
                fill=rgb2hex(cmap(int(v) % 255)),
                outline="black",
            )
        if labels:
            d.text(
                ((x - offset) + (scale / 4), (y - offset) + (scale / 10)),
                str(v)[:4],
                fill="black",
                font=font,
            )
    if axes:
        d.line((0, scale, im.size[0], scale), fill="black")
        d.line((scale, 0, scale, im.size[1]), fill="black")
    if axes_labels:
        for i in range(M.ncols):
            d.text(
                (((i + 1) * scale) + scale / 5, scale / 5),
                str(i),
                fill="black",
                font=font,
            )
        for j in range(M.nrows):
            d.text(
                (scale / 5, ((j + 1) * scale) + scale / 5),
                str(j),
                fill="black",
                font=font,
            )
    if filename is not None:
        im.save(filename + ".png", "PNG")
    return im


def draw_vector(V, column=True, *args, **kwargs):  # pragma: nocover
    from pygraphblas import Matrix

    if column:
        A = Matrix.sparse(V.type, V.size, 1)
        A[:, 0] = V
    else:
        A = Matrix.sparse(V.type, 1, V.size)
        A[0, :] = V

    return draw_matrix(A, *args, **kwargs)


def draw_matrix_op(
    left,
    op,
    right,
    result,
    font_path=None,
    filename=None,
    eqstr="=",
    **kwargs,
):  # pragma: nocover
    scale = kwargs["scale"]
    font = None
    if font_path is not None:
        font = ImageFont.truetype(
            str(font_path / "FantasqueSansMono-Bold.ttf"), int(scale * 0.5)
        )
    else:
        font = ImageFont.load_default()

    kwargs["font_path"] = font_path
    left = draw_matrix(left, **kwargs)
    right = draw_matrix(right, **kwargs)
    result = draw_matrix(result, **kwargs)
    op_width = font.getsize(op)[0]
    spacer = int(scale * 2)
    width = left.size[0] + op_width + spacer + right.size[0] + spacer + result.size[0]
    height = max(op_width + spacer, left.size[1], right.size[1], result.size[1])
    im = Image.new(left.mode, (width, height), color="white")
    d = ImageDraw.Draw(im)
    im.paste(left, (0, 0))
    d.text((left.size[0] + int(spacer / 2), height / 2), op, fill="black", font=font)
    im.paste(right, (left.size[0] + op_width + spacer, 0))
    d.text(
        (left.size[0] + right.size[0] + op_width + (spacer * 1.5), height / 2),
        eqstr,
        fill="black",
        font=font,
    )
    im.paste(result, (left.size[0] + right.size[0] + op_width + spacer + spacer, 0))
    if filename is not None:
        im.save(filename + ".png", "PNG")
    return im


def draw_matrix_layers(layers, **kwargs):  # pragma: nocover
    filename = kwargs.pop("filename", None)
    imgs = [draw_matrix(i, **kwargs) for i in layers]
    widths = [0] + list(accumulate((i.size[0] for i in imgs), operator.add))
    width = sum(widths)
    height = max(i.size[1] for i in imgs)
    im = Image.new(imgs[0].mode, (width, height), color="white")
    for c, i in enumerate(imgs):
        im.paste(i, (widths[c], 0))
    if filename is not None:
        im.save(filename + ".png", "PNG")
    return im


def cy_matrix(M, directed=True):  # pragma: nocover
    nodes = OrderedDict()
    edges = []

    for i, j, v in sorted(M):
        edges.append(
            {"data": {"id": f"{i}:{j}:{v}", "source": str(i), "target": str(j)}}
        )
        if i not in nodes:
            nodes[i] = {"data": {"id": str(i)}}
        if j not in nodes:
            nodes[j] = {"data": {"id": str(j)}}

    return {
        "directed": directed,
        "elements": {"nodes": list(nodes.values()), "edges": edges},
    }


my_style = [
    {
        "selector": "node",
        "style": {
            "background-color": "blue",
            "label": "data(id)",
            "width": 3,
            "height": 3,
            "shape": "circle",
            "color": "#black",
            "font-weight": 3,
            "text-halign": "right",
            "text-valign": "bottom",
            "font-size": 3,
        },
    },
    {
        "selector": "edge",
        "style": {
            "width": 0.1,
            "opacity": 1,
            "line-color": "green",
        },
    },
]


def draw_cy(M, visual_style=my_style, layout_name="cose"):  # pragma: nocover
    from cyjupyter import Cytoscape

    c = Cytoscape(data=cy_matrix(M), visual_style=visual_style, layout_name=layout_name)
    return c


def draw_vis(M, **kwargs):  # pragma: nocover
    from pyvis import network as net

    N = net.Network(**kwargs)
    for i, j, v in M:
        N.add_node(i, i, title=str(i))
        N.add_node(j, j, title=str(j))
        N.add_edge(i, j, value=str(v))
    return N



"""Contains all automatically generated Semirings from CFFI.

This documentation does not show all the semirings in this module
because of the sheer number of them (over 1700).  Please see the
SuiteSparse User Guide for more information on the semirings usable in
The GraphBLAS.

All the standard and extension semirings that comes with SuiteSparse
are represented by objects in this module.  For example
`pygraphblas.semiring.PLUS_TIMES_INT64`.

"""

import sys
import re
import contextvars
from itertools import chain
from collections import defaultdict

from .base import lib, ffi, _check
from .monoid import Monoid
from . import types

current_semiring = contextvars.ContextVar("current_semiring")

__all__ = ["Semiring", "current_semiring"]


class Semiring:

    __slots__ = ("name", "semiring", "token", "pls", "mul", "type")

    def __init__(self, pls, mul, typ, semiring, udt=None):
        self.pls = pls
        self.mul = mul
        self.type = typ
        self.name = "_".join((pls, mul, typ))
        self.semiring = semiring
        self.token = None
        name = pls + "_" + mul
        if udt is None:
            cls = getattr(types, typ)
            setattr(cls, name, self)
            setattr(cls, name.lower(), self)
            types.__pdoc__[f"{typ}.{pls}_{mul}"] = f"UnaryOp {typ}.{pls}_{mul}"

    def __call__(self, A, B, *args, **kwargs):
        from .vector import Vector

        if isinstance(A, Vector):
            op = A.vxm
        elif isinstance(B, Vector):
            op = A.mxv
        else:
            op = A.mxm
        kwargs["semiring"] = self
        return op(B, *args, **kwargs)

    def __enter__(self):
        self.token = current_semiring.set(self)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        current_semiring.reset(self.token)
        return False

    def get_semiring(self, typ=None):
        return self.semiring

    @property
    def ztype(self):
        return types.get_semiring_ztype(self.semiring)

    def print(self, level=2, name="", f=sys.stdout):  # pragma: nocover
        """Print the matrix using `GxB_Matrix_fprint()`, by default to
        `sys.stdout`.

        Level 1: Short description
        Level 2: Short list, short numbers
        Level 3: Long list, short number
        Level 4: Short list, long numbers
        Level 5: Long list, long numbers

        """
        _check(lib.GxB_Semiring_fprint(self.semiring, bytes(name, "utf8"), level, f))


non_boolean_re = re.compile(
    "^(GxB|GrB)_(MIN|MAX|PLUS|TIMES|ANY)_"
    "(FIRST|FIRSTI|FIRSTJ|FIRSTI1|FIRSTJ1|SECOND|SECONDI|SECONDJ|SECONDI1|SECONDJ1|MIN|MAX|PLUS|MINUS|RMINUS|TIMES|DIV|RDIV|ISEQ|ISNE|"
    "ISGT|ISLT|ISGE|ISLE|LOR|LAND|LXOR|PAIR)_"
    "(?:SEMIRING_)?"
    "(UINT8|UINT16|UINT32|UINT64|INT8|INT16|INT32|INT64|FP32|FP64)$"
)

boolean_re = re.compile(
    "^(GxB|GrB)_(LOR|LAND|LXOR|EQ|ANY)_"
    "(EQ|NE|GT|LT|GE|LE)_"
    "(?:SEMIRING_)?"
    "(UINT8|UINT16|UINT32|UINT64|INT8|INT16|INT32|INT64|FP32|FP64|FC32|FC64)$"
)

pure_bool_re = re.compile(
    "^(GxB|GrB)_(LOR|LAND|LXOR|EQ|ANY)_"
    "(FIRST|SECOND|LOR|LAND|LXOR|EQ|GT|LT|GE|LE|PAIR)_"
    "(?:SEMIRING_)?"
    "(BOOL)$"
)

complex_re = re.compile(
    "^(GxB|GrB)_(PLUS|TIMES|ANY)_"
    "(FIRST|SECOND|PLUS|MINUS|RMINUS|TIMES|DIV|RDIV|PAIR)_"
    "(?:SEMIRING_)?"
    "(FC32|FC64)$"
)

bitwise_re = re.compile(
    "^(GxB|GrB)_(BOR|BAND|BXOR|BXNOR)_"
    "(BOR|BAND|BXOR|BXNOR)_"
    "(?:SEMIRING_)?"
    "(UINT8|UINT16|UINT32|UINT64)$"
)


def semiring_group(reg):
    srs = []
    for n in filter(None, [reg.match(i) for i in dir(lib)]):
        prefix, pls, mul, typ = n.groups()
        srs.append(Semiring(pls, mul, typ, getattr(lib, n.string)))
    return srs


def semiring_template(r):  # pragma: nocover
    from .matrix import Matrix

    if r.ztype in (types.FC32, types.FC64):
        return f"Semiring {r.name}"
    A = Matrix.from_lists([0, 0, 1, 1], [0, 1, 0, 1], [1, 0, 1, 0])
    B = A.dup()
    if r.ztype is types.BOOL:
        A = A.pattern()
        B = B.pattern()
    C = A.mxm(B, semiring=r)
    return f"""\
Semiring {r.name}

<table>
<tr>
<td>{A.to_html_table('A')}</td><td> {B.to_html_table('B')}</td><td> {C.to_html_table('A @ B')}</td>
</tr>
</table>
"""


def build_semirings(__pdoc__):
    import tempfile

    this = sys.modules[__name__]
    for r in chain(
        semiring_group(non_boolean_re),
        semiring_group(boolean_re),
        semiring_group(pure_bool_re),
        semiring_group(bitwise_re),
        semiring_group(complex_re),
    ):
        setattr(this, r.name, r)
        pls, mul, typ = r.name.split("_")
        f = tempfile.TemporaryFile()
        r.print(f=f)
        f.seek(0)
        __pdoc__[f"{typ}.{pls}_{mul}"] = f"""```{str(f.read(), 'utf8')}```"""



"""High level wrapper around GraphBLAS Vectors.

"""
import sys
import operator
import weakref
from array import array
from functools import partial
import numpy as np

from .base import (
    lib,
    ffi,
    NULL,
    NoValue,
    _check,
    _error_codes,
    _get_bin_op,
    _get_select_op,
    _build_range,
    GxB_INDEX_MAX,
)
from . import types
from .scalar import Scalar
from .semiring import current_semiring
from .binaryop import current_accum, current_binop, Accum
from .monoid import current_monoid
from . import descriptor
from .descriptor import Descriptor, T1, current_desc


__all__ = ["Vector"]
__pdoc__ = {"Vector.__init__": False}


class Vector:
    """GraphBLAS Sparse Vector

    This is a high-level wrapper around the low-level GrB_Vector type.

    A Vector supports many possible operations according to the
    GraphBLAS API.  Many of those operations have overloaded
    operators.

    Operator | Description | Default
    --- | --- | ---
    v @    A | Vector Vector Multiplication | type default PLUS_TIMES semiring
    v @=   A | In-place Vector Vector Multiplication | type default PLUS_TIMES semiring
    v \\|  w | Vector Union | type default SECOND combiner
    v \\|= w | In-place Vector Union | type default SECOND combiner
    v &    w | Vector Intersection | type default SECOND combiner
    v &=   w | In-place Vector Intersection | type default SECOND combiner
    v +    w | Vector Element-Wise Union | type default PLUS combiner
    v +=   w | In-place Vector Element-Wise Union | type default PLUS combiner
    v -    w | Vector Element-Wise Union | type default MINUS combiner
    v -=   w | In-place Vector Element-Wise Union | type default MINUS combiner
    v *    w | Vector Element-Wise Intersection | type default TIMES combiner
    v *=   w | In-place Vector Element-Wise Intersection | type default TIMES combiner
    v /    w | Vector Element-Wise Intersection | type default DIV combiner
    v /=   w | In-place Vector Element-Wise Intersection | type default DIV combiner
    v ==   w | Compare Element-Wise Union | type default EQ operator
    v !=   w | Compare Element-Wise Union | type default NE operator
    v <    w | Compare Element-Wise Union | type default LT operator
    v >    w | Compare Element-Wise Union | type default GT operator
    v <=   w | Compare Element-Wise Union | type default LE operator
    v >=   w | Compare Element-Wise Union | type default GE operator

    Note that all the above operator syntax is mearly sugar over
    various combinations of calling `Matrix.mxv`, `Vector.vxm`,
    `Vector.eadd`, and `Vector.emult`.

    """

    __slots__ = ("_vector", "type", "_keep_alives")

    def _check(self, res):
        if res != lib.GrB_SUCCESS:
            error_string = ffi.new("char**")
            lib.GrB_Vector_error(error_string, self._vector[0])
            raise _error_codes[res](ffi.string(error_string[0]))

    def __init__(self, vec, typ=None):
        if typ is None:
            new_type = ffi.new("GrB_Type*")
            self._check(lib.GxB_Vector_type(new_type, vec[0]))

            typ = types._gb_type_to_type(new_type[0])

        self._vector = vec
        self.type = typ
        self._keep_alives = weakref.WeakKeyDictionary()

    def __del__(self):
        self._check(lib.GrB_Vector_free(self._vector))

    def __len__(self):
        return self.nvals

    def __iter__(self):
        nvals = self.nvals
        _nvals = ffi.new("GrB_Index[1]", [nvals])
        I = ffi.new("GrB_Index[%s]" % nvals)
        X = ffi.new("%s[%s]" % (self.type._c_type, nvals))
        self._check(self.type._Vector_extractTuples(I, X, _nvals, self._vector[0]))
        return zip(I, X)

    def __getattr__(self, name):
        """Look up operators as attributes for the given object."""
        return partial(getattr(self.type, name), self)

    @property
    def indices(self):
        """cdata array of vector indexes.

        >>> v = Vector.from_1_to_n(3)
        >>> list(v.indices)
        [0, 1, 2]

        """
        nvals = self.nvals
        _nvals = ffi.new("GrB_Index[1]", [nvals])
        I = ffi.new("GrB_Index[%s]" % nvals)
        X = NULL
        self._check(self.type._Vector_extractTuples(I, X, _nvals, self._vector[0]))
        return I

    @property
    def I(self):
        """Iterator over for `Vector.indices`.

        >>> v = Vector.from_1_to_n(3)
        >>> list(v.I)
        [0, 1, 2]

        """
        return iter(self.indices)

    @property
    def npI(self):
        """numpy array over `Vector.indices`.

        >>> v = Vector.from_1_to_n(3)
        >>> v.npI
        array([0, 1, 2], dtype=uint64)

        """
        return np.frombuffer(ffi.buffer(self.indices), dtype=np.uint64)

    @property
    def vals(self):
        """Iterator of vector values.

        >>> v = Vector.from_1_to_n(3)
        >>> list(v.vals)
        [1, 2, 3]

        """
        nvals = self.nvals
        _nvals = ffi.new("GrB_Index[1]", [nvals])
        I = NULL
        V = ffi.new("%s[%s]" % (self.type._c_type, nvals))
        self._check(self.type._Vector_extractTuples(I, V, _nvals, self._vector[0]))
        return V

    @property
    def V(self):
        """Iterator over for `Vector.vals`.

        >>> v = Vector.from_1_to_n(3)
        >>> list(v.V)
        [1, 2, 3]

        """
        return iter(self.vals)

    @property
    def npV(self):
        """numpy array over `Vector.vals`.

        >>> v = Vector.from_1_to_n(3)
        >>> v.npV
        array([1, 2, 3])

        """
        return np.frombuffer(ffi.buffer(self.vals), dtype=self.type._numpy_t)

    def all(self, other, op):
        """Do all elements in self compare True with op to other?

        >>> from . import INT64
        >>> M = Vector.from_lists([0, 1, 2], [1, 2, 3])
        >>> N = Vector.from_lists([0, 1, 2], [1, 2, 3])
        >>> O = Vector.from_lists([0, 1], [1, 2])
        >>> P = Vector.from_lists([0, 1], [1, 2], size=3)
        >>> Q = Vector.from_lists([0, 1, 3], [1, 2, 3])
        >>> assert M.all(N, INT64.eq)
        >>> assert not M.all(N, INT64.gt)
        >>> assert not M.all(O, INT64.eq)
        >>> assert not M.all(P, INT64.eq)
        >>> assert not M.all(Q, INT64.eq)

        """
        if self.size != other.size:
            return False
        if self.nvals != other.nvals:
            return False
        C = self.emult(other, op, cast=types.BOOL)
        if C.nvals != self.nvals:  # pragma: nocover
            return False
        return C.reduce_bool(types.BOOL.land_monoid)

    def iseq(self, other, eq_op=None):
        """Compare two vectors for equality.

        Note to be confused with the `==` operator which does
        element-wise comparison and returns a `Vector`.

        >>> v = Vector.from_lists([0,1], [1, 1])
        >>> w = Vector.from_lists([0,1], [1, 1])
        >>> x = Vector.from_lists([0,1], [1.0, 1.0])
        >>> v.iseq(w)
        True

        >>> v.iseq(w, eq_op=types.UINT64.GE)
        True

        >>> v.iseq(x)
        False
        """
        if self.type != other.type:
            return False
        if eq_op is None:
            eq_op = self.type.EQ
        return self.all(other, eq_op)

    def isne(self, other):
        """Compare two vectors for inequality.
        Note to be confused with the `==` operator which does
        element-wise comparison and returns a `Vector`.

        >>> v = Vector.from_lists([0,1], [1, 1])
        >>> w = Vector.from_lists([0,1], [1, 1])
        >>> v.isne(w)
        False

        """
        return not self.iseq(other)

    @classmethod
    def sparse(cls, typ, size=None):
        """Create an empty Vector from the given type.  If `size` is not
        specified it defaults to `pygraphblas.GxB_INDEX_MAX`.

        >>> v = Vector.sparse(types.INT64, 3)
        >>> v
        <Vector(INT64 size: 3, nvals: 0)>
        >>> v.size
        3
        >>> v = Vector.sparse(types.INT64)
        >>> v
        <Vector(INT64, nvals: 0)>
        >>> v.size == lib.GxB_INDEX_MAX
        True

        """
        if size is None:
            size = GxB_INDEX_MAX
        new_vec = ffi.new("GrB_Vector*")
        _check(lib.GrB_Vector_new(new_vec, typ._gb_type, size))
        return cls(new_vec, typ)

    @classmethod
    def from_lists(cls, I, V, size=None, typ=None):
        """Create a new vector from the given lists of indices and values.  If
        size is not provided, it is computed from the max values of
        the provides size indices.

        If the second argument is a scalar value, an "iso" vector is
        created where all values equal that scalar.

        >>> v = Vector.from_lists([0, 1, 2], [1, 2, 3])
        >>> w = Vector.from_lists([0, 1, 2], True)
        >>> assert not v.iseq(w)
        >>> assert v.pattern().iseq(w)
        """
        if isinstance(V, (bool, int, float)):
            V = [V] * len(I)
        assert len(I) == len(V)
        assert len(I) > 0  # must be non empty
        if not size:
            size = max(I) + 1
        # TODO option to use ffi and GrB_Vector_build
        if typ is None:
            typ = types._gb_from_type(type(V[0]))
        m = cls.sparse(typ, size)
        for i, v in zip(I, V):
            m[i] = v
        return m

    @classmethod
    def from_list(cls, I):
        """Create a new dense vector from the given lists of values."""
        size = len(I)
        assert size > 0
        # TODO use ffi and GrB_Vector_build
        m = cls.sparse(types._gb_from_type(type(I[0])), size)
        for i, v in enumerate(I):
            m[i] = v
        return m

    @classmethod
    def from_1_to_n(cls, n):
        """Generate a vector from 1 to n.

        >>> v = Vector.from_1_to_n(3)
        >>> print(v)
        0| 1
        1| 2
        2| 3
        """
        v = cls.sparse(types.INT64, n)
        for i in range(n):
            v[i] = i + 1
        return v

    def dup(self):
        """Create an duplicate Vector from the given argument.

        >>> v = Vector.from_1_to_n(3)
        >>> w = v.dup()
        >>> w is not v
        True
        >>> w.iseq(v)
        True
        >>> print(w)
        0| 1
        1| 2
        2| 3
        """
        new_vec = ffi.new("GrB_Vector*")
        self._check(lib.GrB_Vector_dup(new_vec, self._vector[0]))
        return self.__class__(new_vec, self.type)

    @property
    def hyper_switch(self):  # pragma: nocover
        """Get the hyper_switch threshold. (See SuiteSparse User Guide)"""
        switch = ffi.new("double*")
        self._check(
            lib.GxB_Vector_Option_get(self._vector[0], lib.GxB_HYPER_SWITCH, switch)
        )
        return switch[0]

    @hyper_switch.setter
    def hyper_switch(self, switch):  # pragma: nocover
        """Set the hyper_switch threshold. (See SuiteSparse User Guide)"""
        switch = ffi.cast("double", switch)
        self._check(
            lib.GxB_Vector_Option_set(self._vector[0], lib.GxB_HYPER_SWITCH, switch)
        )

    @property
    def sparsity(self):  # pragma: nocover
        """Get Vector sparsity control. (See SuiteSparse User Guide)"""
        sparsity = ffi.new("int*")
        self._check(
            lib.GxB_Vector_Option_get(
                self._vector[0], lib.GxB_SPARSITY_CONTROL, sparsity
            )
        )
        return sparsity[0]

    @sparsity.setter
    def sparsity(self, sparsity):  # pragma: nocover
        """Set Vector sparsity control. (See SuiteSparse User Guide)"""
        sparsity = ffi.cast("int", sparsity)
        self._check(
            lib.GxB_Vector_Option_set(
                self._vector[0], lib.GxB_SPARSITY_CONTROL, sparsity
            )
        )

    @property
    def sparsity_status(self):  # pragma: nocover
        """Get Vector sparsity status. (See SuiteSparse User Guide)"""
        status = ffi.new("int*")
        self._check(
            lib.GxB_Vector_Option_get(self._vector[0], lib.GxB_SPARSITY_STATUS, status)
        )
        return status[0]

    @classmethod
    def dense(cls, typ, size=None, fill=None):
        """Return a dense vector of `typ` and `size`.  If `fill` is provided,
        use that value otherwise use `self.type.default_zero`

        >>> print(Vector.dense(types.FP32, 3))
        0|0.0
        1|0.0
        2|0.0
        >>> print(Vector.dense(types.FP32, 3, fill=42.0))
        0|42.0
        1|42.0
        2|42.0

        """
        v = cls.sparse(typ, size)
        if fill is None:
            fill = v.type.default_zero
        v[:] = fill
        return v

    @classmethod
    def iso(cls, value, size=GxB_INDEX_MAX):
        """Build an "iso" vector from a scalar value.

        This is similar to `Vector.dense` but infers the type of the
        new Matrix from the provided vbalue.

        >>> v = Vector.iso(3)
        >>> assert v[42] == 3
        """
        typ = types._gb_from_type(type(value))
        return cls.dense(typ, size, value)

    def to_lists(self):
        """Extract the indices and values of the Vector as 2 lists.

        >>> Vector.from_1_to_n(3).to_lists()
        [[0, 1, 2], [1, 2, 3]]

        """
        I = ffi.new("GrB_Index[]", self.nvals)
        V = self.type._ffi.new(self.type._c_type + "[]", self.nvals)
        n = ffi.new("GrB_Index*")
        n[0] = self.nvals
        self._check(self.type._Vector_extractTuples(I, V, n, self._vector[0]))
        return [list(I), list(map(self.type._to_value, V))]

    def to_arrays(self):
        """Return as python `array` objects.

        >>> Vector.from_1_to_n(3).to_arrays()
        (array('L', [0, 1, 2]), array('q', [1, 2, 3]))

        """
        if self.type._typecode is None:
            raise TypeError("This matrix has no array typecode.")
        nvals = self.nvals
        _nvals = ffi.new("GrB_Index[1]", [nvals])
        I = ffi.new("GrB_Index[%s]" % nvals)
        X = self.type._ffi.new("%s[%s]" % (self.type._c_type, nvals))
        self._check(self.type._Vector_extractTuples(I, X, _nvals, self._vector[0]))
        return array("L", I), array(self.type._typecode, X)

    @property
    def size(self):
        """Return the size of the vector.

        >>> Vector.from_1_to_n(3).size
        3

        """
        n = ffi.new("GrB_Index*")
        self._check(lib.GrB_Vector_size(n, self._vector[0]))
        return n[0]

    @property
    def nvals(self):
        """Return the number of values in the vector.

        >>> v = Vector.from_1_to_n(3)
        >>> v.nvals
        3
        >>> v.clear()
        >>> v.nvals
        0

        """
        n = ffi.new("GrB_Index*")
        self._check(lib.GrB_Vector_nvals(n, self._vector[0]))
        return n[0]

    @property
    def memory_usage(self):
        """Returns the memory usage of the Vector.

        >>> v = Vector.from_lists([0, 1, 2], [1, 2, 0])
        >>> assert v.memory_usage > 0
        """
        n = ffi.new("size_t*")
        self._check(lib.GxB_Vector_memoryUsage(n, self._vector[0]))
        return n[0]

    @property
    def gb_type(self):
        """Return the GraphBLAS low-level type object of the Vector."""
        return self.type._gb_type

    def _full(self):
        B = self.__class__.sparse(self.type, self.size)

        self._check(
            self.type._Vector_assignScalar(
                B._vector[0], NULL, NULL, self.type.default_one, lib.GrB_ALL, 0, NULL
            )
        )
        return self.eadd(B, self.type.FIRST)

    def _compare(self, other, op, strop):
        C = self.__class__.sparse(types.BOOL, self.size)
        if isinstance(other, (bool, int, float, complex)):
            if op(other, 0):
                B = self.__class__.dup(self)
                B[:] = other
                self.emult(B, strop, out=C)
                return C
            else:
                self.select(strop, other).apply(types.BOOL.ONE, out=C)
                return C
        elif isinstance(other, Vector):
            A = self._full()
            B = other._full()
            A.emult(B, strop, out=C)
            return C
        else:
            raise NotImplementedError

    def __gt__(self, other):
        return self._compare(other, operator.gt, ">")

    def __lt__(self, other):
        return self._compare(other, operator.lt, "<")

    def __ge__(self, other):
        return self._compare(other, operator.ge, ">=")

    def __le__(self, other):
        return self._compare(other, operator.le, "<=")

    def __eq__(self, other):
        return self._compare(other, operator.eq, "==")

    def __ne__(self, other):
        return self._compare(other, operator.ne, "!=")

    def eadd(
        self,
        other,
        add_op=None,
        cast=None,
        out=None,
        mask=None,
        accum=None,
        desc=None,
    ):
        """Element-wise addition with other vector.

        Element-wise addition applies a binary operator element-wise
        on two vectors `v` and `w`, for all entries that appear in the
        set union of the patterns of `A` and `B`.

        The only difference between element-wise multiplication and
        addition is the pattern of the result, and what happens to
        entries outside the intersection. With multiplication the
        pattern of T is the intersection; with addition it is the set
        union. Entries outside the set intersection are dropped for
        multiplication, and kept for addition; in both cases the
        operator is only applied to those (and only those) entries in
        the intersection. Any binary operator can be used
        interchangeably for either operation.

        >>> I = [0, 0, 1, 1, 2, 3, 3, 4, 5, 6, 6, 6]
        >>> V = list(range(len(I)))
        >>> v = Vector.from_lists(I, V, 7)

        >>> w = Vector.from_lists(
        ...    [0, 1, 4, 6],
        ...    [9, 1, 4, 7], 7)

        >>> print(v.eadd(w))
        0|10
        1| 4
        2| 4
        3| 6
        4|11
        5| 8
        6|18

        This can also be accomplished with the `+` operators:

        >>> print(v + w)
        0|10
        1| 4
        2| 4
        3| 6
        4|11
        5| 8
        6|18

        The combining operator used can be provided either as a
        context manager or passed to `mxv` as the `add_op` argument.

        >>> with types.INT64.MIN:
        ...     print(v + w)
        0| 1
        1| 1
        2| 4
        3| 6
        4| 4
        5| 8
        6| 7

        The following operators default to use `eadd`:

        Operator | Description | Default
        --- | --- | ---
        v \\|  w | Vector Union | type default SECOND combiner
        v \\|= w | In-place Vector Union | type default SECOND combiner
        v +    w | Vector Element-Wise Union | type default PLUS combiner
        v +=   w | In-place Vector Element-Wise Union | type default PLUS combiner
        v -    w | Vector Element-Wise Union | type default MINUS combiner
        v -=   w | In-place Vector Element-Wise Union | type default MINUS combiner

        """
        if add_op is None:
            add_op = current_binop.get(NULL)

        mask, accum, desc = self._get_args(mask, accum, desc)

        if out is None:
            typ = cast or types.promote(self.type, other.type)
            _out = ffi.new("GrB_Vector*")
            self._check(lib.GrB_Vector_new(_out, typ._gb_type, self.size))
            out = self.__class__(_out, typ)

        if add_op is NULL:
            add_op = out.type._default_addop()
        add_op = add_op.get_binaryop(self.type, other.type)
        self._check(
            lib.GrB_Vector_eWiseAdd_BinaryOp(
                out._vector[0],
                mask,
                accum,
                add_op,
                self._vector[0],
                other._vector[0],
                desc,
            )
        )
        return out

    def emult(
        self,
        other,
        mult_op=None,
        cast=None,
        out=None,
        mask=None,
        accum=None,
        desc=None,
    ):
        """Element-wise multiplication with other vector.

        Element-wise multiplication applies a binary operator
        element-wise on two vectors A and B, for all entries that
        appear in the set intersection of the patterns of A and B.

        The only difference between element-wise multiplication and
        addition is the pattern of the result, and what happens to
        entries outside the intersection. With multiplication the
        pattern of T is the intersection; with addition it is the set
        union. Entries outside the set intersection are dropped for
        multiplication, and kept for addition; in both cases the
        operator is only applied to those (and only those) entries in
        the intersection. Any binary operator can be used
        interchangeably for either operation.

        >>> I = [0, 0, 1, 1, 2, 3, 3, 4, 5, 6, 6, 6]
        >>> V = list(range(len(I)))
        >>> v = Vector.from_lists(I, V, 7)

        >>> w = Vector.from_lists(
        ...    [0, 1, 4, 6],
        ...    [9, 1, 4, 7], 7)

        >>> print(v.emult(w))
        0| 9
        1| 3
        2|
        3|
        4|28
        5|
        6|77

        This can also be accomplished with the `+` operators:

        >>> print(v * w)
        0| 9
        1| 3
        2|
        3|
        4|28
        5|
        6|77

        The combining operator used can be provided either as a
        context manager or passed to `mxv` as the `add_op` argument.

        >>> with types.INT64.MAX:
        ...     print(v * w)
        0| 9
        1| 3
        2|
        3|
        4|28
        5|
        6|77

        """
        if mult_op is None:
            mult_op = current_binop.get(NULL)
        elif isinstance(mult_op, str):
            mult_op = _get_bin_op(mult_op, self.type)

        mask, accum, desc = self._get_args(mask, accum, desc)
        if out is None:
            typ = cast or types.promote(self.type, other.type)
            _out = ffi.new("GrB_Vector*")
            self._check(lib.GrB_Vector_new(_out, typ._gb_type, self.size))
            out = self.__class__(_out, typ)

        if mult_op is NULL:
            mult_op = out.type._default_multop()

        mult_op = mult_op.get_binaryop(self.type, other.type)
        self._check(
            lib.GrB_Vector_eWiseMult_BinaryOp(
                out._vector[0],
                mask,
                accum,
                mult_op,
                self._vector[0],
                other._vector[0],
                desc,
            )
        )
        return out

    def vxm(
        self,
        other,
        cast=None,
        out=None,
        semiring=None,
        mask=None,
        accum=None,
        desc=None,
    ):
        """Vector-Matrix multiply.


        Multiply this row vector by `other` matrix "on the left".  For
        column matrix/vector multiplication "on the right" see
        `Matrix.mxv`.

        `vxm` can also be called directly or with the `@` operator:

        >>> from . import Matrix, INT64
        >>> M = Matrix.from_lists([0, 1, 2], [1, 2, 0], [1, 2, 3])
        >>> v = Vector.from_lists([0, 1, 2], [2, 3, 4])
        >>> o = v.vxm(M)
        >>> print(o)
        0|12
        1| 2
        2| 6
        >>> o = v @ M
        >>> print(o)
        0|12
        1| 2
        2| 6

        By default, `mxv` and `@` create a new result matrix of the
        correct type and dimensions if one is not provided.  If you
        want to provide your own matrix to put the result in, you can
        pass it in the `out` parameter.  This is useful for
        accumulating results into a single matrix with minimal
        copying.  This is also supported by the `@=` syntax:

        >>> o = v.dup()
        >>> v.vxm(M, accum=INT64.plus, out=o) is o
        True
        >>> print(o)
        0|14
        1| 5
        2|10
        >>> o = v.dup()
        >>> with Accum(INT64.min):
        ...     o @= M
        >>> print(o)
        0| 2
        1| 2
        2| 4

        The default semiring depends on the infered result type.  In
        the case of numbers, the default semiring is `PLUS_TIMES`.  In
        the case of type `BOOL`, it is `BOOL.lor_land`.

        >>> o = v.vxm(M, semiring=INT64.min_plus)
        >>> print(o)
        0| 7
        1| 3
        2| 5

        An explicit semiring can be passed to the method or provided
        with a context manager:

        >>> with INT64.min_plus:
        ...     o = v @ M
        >>> print(o)
        0| 7
        1| 3
        2| 5

        Or the semiring can be accessed via an attribute on the
        vector:

        >>> o = v.min_plus(M)
        >>> print(o)
        0| 7
        1| 3
        2| 5

        Descriptors and accumulators can also be provided as an
        argument or a context manager:

        >>> o = v.vxm(M, desc=descriptor.T0)
        >>> print(o)
        0|12
        1| 2
        2| 6
        >>> with descriptor.T0:
        ...     o = v @ M
        >>> print(o)
        0|12
        1| 2
        2| 6
        >>> del o[1]
        >>> o = v.vxm(M, mask=o)
        >>> print(o)
        0|12
        1|
        2| 6

        """

        if semiring is None:
            semiring = current_semiring.get(NULL)

        if out is None:
            new_dimension = other.nrows if T1 in (desc or ()) else other.ncols
            if semiring is not NULL:
                typ = semiring.ztype
            else:
                typ = cast or types.promote(self.type, other.type)
            out = Vector.sparse(typ, new_dimension)
        else:
            typ = out.type

        if semiring is NULL:
            semiring = out.type._default_semiring()

        semiring = semiring.get_semiring()
        mask, accum, desc = self._get_args(mask, accum, desc)
        self._check(
            lib.GrB_vxm(
                out._vector[0],
                mask,
                accum,
                semiring,
                self._vector[0],
                other._matrix[0],
                desc,
            )
        )
        return out

    def __matmul__(self, other):
        return self.vxm(other)

    def __imatmul__(self, other):
        return self.vxm(other, out=self)

    def __and__(self, other):
        return self.emult(other)

    def __iand__(self, other):
        return self.emult(other, out=self)

    def __or__(self, other):
        return self.eadd(other)

    def __ior__(self, other):
        return self.eadd(other, out=self)

    def __add__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.PLUS, other)
        return self.eadd(other)

    def __radd__(self, other):
        if not isinstance(other, Vector):
            return self.apply_first(other, self.type.PLUS)
        return other.eadd(self)  # pragma: nocover

    def __iadd__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.PLUS, other, out=self)
        return self.eadd(other, out=self)

    def __sub__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.MINUS, other)
        return self.eadd(other, self.type.MINUS)

    def __rsub__(self, other):
        if not isinstance(other, Vector):
            return self.apply_first(other, self.type.MINUS)
        return other.eadd(self, self.type.MINUS)  # pragma: nocover

    def __isub__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.MINUS, other)
        return other.eadd(self, self.type.MINUS, out=self)

    def __mul__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.TIMES, other)
        return self.emult(other, self.type.TIMES)

    def __rmul__(self, other):
        if not isinstance(other, Vector):
            return self.apply_first(other, self.type.TIMES)
        return other.emult(self, add_op=self.type.TIMES)  # pragma: nocover

    def __imul__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.TIMES, other, out=self)
        return other.emult(self, self.type.TIMES, out=self)

    def __truediv__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.DIV, other)
        return self.emult(other, self.type.DIV)

    def __rtruediv__(self, other):
        if not isinstance(other, Vector):
            return self.apply_first(other, self.type.DIV)
        return other.emult(self, self.type.DIV)  # pragma: nocover

    def __itruediv__(self, other):
        if not isinstance(other, Vector):
            return self.apply_second(self.type.DIV, other, out=self)
        return other.emult(self, self.type.DIV, out=self)

    def __invert__(self):
        return self.apply(self.type.MINV)

    def __neg__(self):
        return self.apply(self.type.AINV)

    def __abs__(self):
        return self.apply(self.type.ABS)

    def clear(self):
        """Clear this vector removing all entries."""
        self._check(lib.GrB_Vector_clear(self._vector[0]))

    def resize(self, size=lib.GxB_INDEX_MAX):
        """Resize the vector.  If the dimensions decrease, entries that fall
        outside the resized vector are deleted.

        >>> v = Vector.dense(types.UINT8, 2)
        >>> v.resize(3)
        >>> print(v)
        0| 0
        1| 0
        2|

        """
        self._check(lib.GrB_Vector_resize(self._vector[0], size))

    def _get_args(self, mask=None, accum=None, desc=None):
        if accum is None:
            accum = current_accum.get(NULL)

        if accum is not NULL:
            accum = accum.get_binaryop(self.type)

        if desc is None:
            desc = current_desc.get(NULL)

        if desc is not NULL:
            desc = desc.get_desc()

        if mask is None:
            mask = NULL

        if isinstance(mask, Vector):
            mask = mask._vector[0]
        return mask, accum, desc

    def reduce_bool(self, mon=None, mask=None, accum=None, desc=None):
        """Reduce vector to a boolean.

        >>> v = Vector.from_lists([0, 1], [True, False])
        >>> v.reduce_bool()
        True
        >>> v[0] = False
        >>> v.reduce_bool()
        False
        >>> v[1] = True
        >>> v.reduce_bool(types.BOOL.LAND_MONOID)
        False

        """
        if mon is None:
            mon = current_monoid.get(types.BOOL.LOR_MONOID)
        mon = mon.get_monoid(self.type)
        mask, accum, desc = self._get_args(mask, accum, desc)
        result = ffi.new("_Bool*")
        self._check(
            lib.GrB_Vector_reduce_BOOL(result, accum, mon, self._vector[0], desc)
        )
        return result[0]

    def reduce_int(self, mon=None, mask=None, accum=None, desc=None):
        """Reduce vector to a integer.

        >>> v = Vector.from_lists([0, 1], [1, 1])
        >>> v.reduce_int()
        2
        >>> v[0] = 0
        >>> v.reduce_int()
        1
        >>> v[1] = 2
        >>> v.reduce_int(types.INT64.MIN_MONOID)
        0

        """
        if mon is None:
            mon = current_monoid.get(types.INT64.PLUS_MONOID)
        mon = mon.get_monoid(self.type)
        mask, accum, desc = self._get_args(mask, accum, desc)
        result = ffi.new("int64_t*")
        self._check(
            lib.GrB_Vector_reduce_INT64(result, accum, mon, self._vector[0], desc)
        )
        return result[0]

    def reduce_float(self, mon=None, mask=None, accum=None, desc=None):
        """Reduce vector to a float.

        >>> v = Vector.from_lists([0, 1], [1.2, 1.1])
        >>> v.reduce_float()
        2.3
        >>> v[0] = 0
        >>> v.reduce_float()
        1.1
        >>> v[1] = 2.2
        >>> v.reduce_float(types.FP64.MIN_MONOID)
        0.0

        """
        if mon is None:
            mon = current_monoid.get(types.FP64.PLUS_MONOID)
        mon = mon.get_monoid(self.type)
        mask, accum, desc = self._get_args(mask, accum, desc)
        result = ffi.new("double*")
        self._check(
            lib.GrB_Vector_reduce_FP64(result, accum, mon, self._vector[0], desc)
        )
        return result[0]

    def max(self):
        """Return the max of the vector.

        >>> M = Vector.from_lists([0, 1, 2], [False, False, False])
        >>> M.max()
        False
        >>> M = Vector.from_lists([0, 1, 2], [False, False, True])
        >>> M.max()
        True
        >>> M = Vector.from_lists([0, 1, 2], [-42, 0, 149])
        >>> M.max()
        149
        >>> M = Vector.from_lists([0, 1, 2], [-42.0, 0.0, 149.0])
        >>> M.max()
        149.0
        >>> M = Vector.from_lists([0], [1j])
        >>> M.max()
        Traceback (most recent call last):
        ...
        TypeError: Un-maxable type
        """
        if self.type == types.BOOL:
            return self.reduce_bool(self.type.LOR_MONOID)
        if self.type in types._int_types:
            return self.reduce_int(self.type.MAX_MONOID)
        if self.type in types._float_types:
            return self.reduce_float(self.type.MAX_MONOID)
        raise TypeError("Un-maxable type")

    def min(self):
        """Return the min of the vector.

        >>> M = Vector.from_lists([0, 1, 2], [True, True, True])
        >>> M.min()
        True
        >>> M = Vector.from_lists([0, 1, 2], [False, True, True])
        >>> M.min()
        False
        >>> M = Vector.from_lists([0, 1, 2], [-42, 0, 149])
        >>> M.min()
        -42
        >>> M = Vector.from_lists([0, 1, 2], [-42.0, 0.0, 149.0])
        >>> M.min()
        -42.0
        >>> M = Vector.from_lists([0], [1j])
        >>> M.min()
        Traceback (most recent call last):
        ...
        TypeError: Un-minable type
        """
        if self.type == types.BOOL:
            return self.reduce_bool(self.type.LAND_MONOID)
        if self.type in types._int_types:
            return self.reduce_int(self.type.MIN_MONOID)
        if self.type in types._float_types:
            return self.reduce_float(self.type.MIN_MONOID)
        raise TypeError("Un-minable type")

    def apply(self, op, out=None, mask=None, accum=None, desc=None):
        """Apply Unary op to vector elements.
        >>> from . import UINT64
        >>> v = Vector.from_lists([0,1], [1, 1])
        >>> print(v.apply(UINT64.ainv))
        0|-1
        1|-1

        Unary operators can also be accessed by atribute name on
        vectors they are applied to:

        >>> print(v.ainv())
        0|-1
        1|-1

        """
        if out is None:
            out = Vector.sparse(self.type, self.size)

        op = op.get_unaryop(self)
        mask, accum, desc = self._get_args(mask, accum, desc)
        self._check(
            lib.GrB_Vector_apply(out._vector[0], mask, accum, op, self._vector[0], desc)
        )
        return out

    def apply_first(self, first, op, out=None, mask=None, accum=None, desc=None):
        """Apply a binary operator to the entries in a vector, binding the first input
        to a scalar first.


        >>> v = Vector.from_lists([0,1], [1, 1])
        >>> print(v.apply_first(3, types.UINT64.PLUS))
        0| 4
        1| 4
        >>> w = Vector.sparse(v.type, v.size)
        >>> v.apply_first(3, types.UINT64.PLUS, out=w) is w
        True

        """
        if out is None:
            out = self.__class__.sparse(self.type, self.size)

        op = op.get_binaryop(self)
        mask, accum, desc = self._get_args(mask, accum, desc)
        if isinstance(first, Scalar):
            f = lib.GxB_Vector_apply_BinaryOp1st
            first = first._scalar[0]
        else:
            f = self.type._Vector_apply_BinaryOp1st
        self._check(f(out._vector[0], mask, accum, op, first, self._vector[0], desc))
        return out

    def apply_second(self, op, second, out=None, mask=None, accum=None, desc=None):
        """Apply a binary operator to the entries in a vector, binding the second input
        to a scalar second.

        >>> v = Vector.from_lists([0,1], [1, 1])
        >>> print(v.apply_second(types.UINT64.PLUS, 3))
        0| 4
        1| 4
        >>> w = Vector.sparse(v.type, v.size)
        >>> v.apply_second(types.UINT64.PLUS, 3, out=w) is w
        True
        >>> u = Vector.from_lists([0,1], [1.1, 2.2])
        >>> u.apply_second(u.type.TIMES, 3.3, out=u) is u
        True
        >>> u = Vector.from_lists([0,1], [1.1, 2.2])
        >>> print(u * 3)
        0|3.3
        1|6.6
        >>> x = Vector.from_lists([0,1], [1.1, 2.2])
        >>> x *= 3.0
        >>> print(x)
        0|3.3
        1|6.6

        """
        if out is None:
            out = self.__class__.sparse(self.type, self.size)

        op = op.get_binaryop(self)
        mask, accum, desc = self._get_args(mask, accum, desc)
        if isinstance(second, Scalar):
            f = lib.GxB_Vector_apply_BinaryOp2nd
            second = second._scalar[0]
        else:
            f = self.type._Vector_apply_BinaryOp2nd

        self._check(f(out._vector[0], mask, accum, op, self._vector[0], second, desc))
        return out

    def select(self, op, thunk=None, out=None, mask=None, accum=None, desc=None):
        """Select elements that match the given select operation condition.
        See `Matrix.select` for possible operators.

        >>> v = Vector.from_lists([0,1], [1, 0])
        >>> print(v.select('>', 0))
        0| 1
        1|

        >>> w = Vector.sparse(types.UINT8, 2)
        >>> v.select('>', 0, out=w) is w
        True

        `min` and `max` selectors can be shortcuts for selecting all
        elements that equal the min or max reduction of all elements.

        >>> print(v.select('min'))
        0|
        1| 0
        >>> print(v.select('max'))
        0| 1
        1|

        """
        if out is None:
            out = Vector.sparse(self.type, self.size)
        if isinstance(op, str):
            if op == "min":
                op = lib.GxB_EQ_THUNK
                thunk = self.min()
            elif op == "max":
                op = lib.GxB_EQ_THUNK
                thunk = self.max()
            else:
                op = _get_select_op(op)

        if thunk is None:
            thunk = NULL
        if isinstance(thunk, (bool, int, float, complex)):
            thunk = Scalar.from_value(thunk)
        if isinstance(thunk, Scalar):
            self._keep_alives[self._vector] = thunk
            thunk = thunk._scalar[0]

        mask, accum, desc = self._get_args(mask, accum, desc)
        self._check(
            lib.GxB_Vector_select(
                out._vector[0], mask, accum, op, self._vector[0], thunk, desc
            )
        )
        return out

    def pattern(self, typ=types.BOOL):
        """Return the pattern of the vector, this is a boolean Vector where
        every present value in this vector is set to True.

        """
        result = Vector.sparse(typ, self.size)
        self.apply(types.BOOL.ONE, out=result)
        return result

    @property
    def S(self):
        """Return the vector "structure".  This is the same as calling
        `Vector.pattern()` with no arguments.

        >>> v = Vector.from_lists([0, 1, 2], [1, 2, 3])
        >>> assert v.S == v.pattern()

        """
        return self.pattern()

    def nonzero(self):
        """Select vector of nonzero entries."""
        return self.select(lib.GxB_NONZERO)

    def __setitem__(self, index, value):
        if isinstance(index, int):
            val = self.type._from_value(value)
            self._check(self.type._Vector_setElement(self._vector[0], val, index))
            return

        if isinstance(index, slice):
            if isinstance(value, Vector):
                self.assign(value, index)
                return
            if isinstance(value, (bool, int, float, complex)):
                self.assign_scalar(value, index)
                return

        if isinstance(index, Vector):
            mask = index._vector[0]
            index = slice(None, None, None)
            if isinstance(value, Vector):
                self.assign(value, index, mask=mask)
                return
            self.assign_scalar(value, index, mask=mask)
            return
        raise TypeError("Unknown index")

    def assign(self, value, index=None, mask=None, accum=None, desc=None):
        """Assign vector to vector.

        >>> v = Vector.sparse(types.INT8, 3)
        >>> w = Vector.from_1_to_n(3)
        >>> v[:] = w
        >>> print(v)
        0| 1
        1| 2
        2| 3

        If the index is another vector it is used as an assignment
        mask:

        >>> v.clear()
        >>> m = Vector.sparse(types.BOOL, 3)
        >>> m[1] = True
        >>> v[m] = w
        >>> print(v)
        0|
        1| 2
        2|
        >>> v.clear()
        >>> m = Vector.sparse(types.BOOL, 3)
        >>> m[1] = True
        >>> v[m] = 3
        >>> print(v)
        0|
        1| 3
        2|

        """
        mask, accum, desc = self._get_args(mask, accum, desc)
        I, ni, size = _build_range(index, self.size - 1)
        self._check(
            lib.GrB_Vector_assign(
                self._vector[0], mask, accum, value._vector[0], I, ni, desc
            )
        )

    def assign_scalar(self, value, index=None, mask=None, accum=None, desc=None):
        """Assign scalar to vector.

        >>> v = Vector.sparse(types.INT8, 3)
        >>> v[:] = 2
        >>> print(v)
        0| 2
        1| 2
        2| 2

        If the index is another vector it is used as an assignment
        mask:

        >>> v.clear()
        >>> m = Vector.sparse(types.BOOL, 3)
        >>> m[1] = True
        >>> v[m] = 3
        >>> print(v)
        0|
        1| 3
        2|

        """
        mask, accum, desc = self._get_args(mask, accum, desc)
        scalar_type = types._gb_from_type(type(value))
        I, ni, size = _build_range(index, self.size - 1)
        self._check(
            scalar_type._Vector_assignScalar(
                self._vector[0], mask, accum, value, I, ni, desc
            )
        )

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.extract_element(index)
        else:
            return self.extract(index)

    def __delitem__(self, index):
        if not isinstance(index, int):
            raise TypeError(
                "__delitem__ currently only supports single element removal"
            )
        self._check(lib.GrB_Vector_removeElement(self._vector[0], index))

    def extract_element(self, index):
        """Extract element from vector."""
        result = self.type._ffi.new(self.type._ptr)
        self._check(
            self.type._Vector_extractElement(
                result, self._vector[0], ffi.cast("GrB_Index", index)
            )
        )
        return self.type._to_value(result[0])

    def extract(self, index, mask=None, accum=None, desc=None):
        """Extract subvector from vector."""
        mask, accum, desc = self._get_args(mask, accum, desc)
        if isinstance(index, Vector):
            mask = index._vector[0]
            index = slice(None, None, None)
        I, ni, size = _build_range(index, self.size - 1)
        if size is None:
            size = self.size
        result = Vector.sparse(self.type, size)
        self._check(
            lib.GrB_Vector_extract(
                result._vector[0], mask, accum, self._vector[0], I, ni, desc
            )
        )
        return result

    def __contains__(self, index):
        try:
            v = self[index]
            return True
        except NoValue:
            return False

    def get(self, i, default=None):
        """Get element at `i` or return `default` if not present.


        >>> M = Vector.from_lists([1, 2], [42, 149])
        >>> M.get(1)
        42
        >>> M.get(0) is None
        True
        >>> M.get(0, 'foo')
        'foo'

        """
        try:
            return self[i]
        except NoValue:
            return default

    def wait(self):
        """Wait for vector to complete."""
        self._check(lib.GrB_Vector_wait(self._vector))

    def to_string(self, format_string="{:>%s}", width=2, prec=3, empty_char=""):
        """Return string representation of vector."""
        format_string = format_string % width
        result = ""
        for row in range(self.size):
            value = self.get(row, empty_char)
            result += str(row) + "|"
            result += format_string.format(
                self.type.format_value(value, width, prec)
            ).rstrip()
            if row < self.size - 1:
                result += "\n"
        return result

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        tname = self.type.__name__
        if self.size == lib.GxB_INDEX_MAX:
            return f"<Vector({tname}, nvals: {self.nvals})>"

        return f"<Vector({tname} size: {self.size}, nvals: {self.nvals})>"

    def print(self, level=2, name="A", f=sys.stdout):  # pragma: nocover
        """Print the matrix using `GxB_Matrix_fprint()`, by default to
        `sys.stdout`..

        Level 1: Short description
        Level 2: Short list, short numbers
        Level 3: Long list, short number
        Level 4: Short list, long numbers
        Level 5: Long list, long numbers

        """
        self._check(
            lib.GxB_Vector_fprint(self._vector[0], bytes(name, "utf8"), level, f)
        )



"""Contains all automatically generated UnaryOps from CFFI.

"""

__all__ = ["UnaryOp", "unary_op"]

import re, sys
from itertools import chain
from textwrap import dedent
from cffi import FFI
import numba
from numba import cfunc, jit, carray
from numba.core.typing import cffi_utils as cffi_support
import contextvars
from collections import defaultdict

from .base import lib, ffi as core_ffi, _check
from . import types


class UnaryOp:
    """Wrapper around GrB_UnaryOpl"""

    __slots__ = ("name", "unaryop", "ffi", "token")

    def __init__(self, name, typ, op):
        self.name = "_".join((name, typ))
        self.unaryop = op
        self.token = None
        cls = getattr(types, typ)
        setattr(cls, name, self)
        setattr(cls, name.lower(), self)
        types.__pdoc__[f"{typ}.{name}"] = f"UnaryOp {typ}.{name}"

    def __call__(self, A, *args, **kwargs):
        return A.apply(self, *args, **kwargs)

    def get_unaryop(self, operand1=None):
        return self.unaryop

    def print(self, level=2, name="", f=sys.stdout):  # pragma: nocover
        """Print the matrix using `GxB_Matrix_fprint()`, by default to
        `sys.stdout`.

        Level 1: Short description
        Level 2: Short list, short numbers
        Level 3: Long list, short number
        Level 4: Short list, long numbers
        Level 5: Long list, long numbers

        """
        _check(lib.GxB_UnaryOp_fprint(self.unaryop, bytes(name, "utf8"), level, f))


uop_re = re.compile(
    "^(GrB|GxB)_(ONE|ABS|SQRT|LOG|EXP|LOG2|SIN|COS|TAN|ACOS|ASIN|ATAN|SINH|"
    "POSITIONI|POSITIONI1|POSITIONJ|POSITIONJ1|"
    "COSH|TANH|ACOSH|ASINH|ATANH|SIGNUM|CEIL|FLOOR|ROUND|TRUNC|EXP2|EXPM1|"
    "LOG12|LOG1P|LGAMMA|TGAMMA|ERF|ERFC|FREXPX|FREXPE|CONJ|CREAL|CIMAG|CARG|"
    "IDENTITY|AINV|MINV|LNOT|ONE|ABS|ISINF|ISNAN|ISFINITE)_"
    "(BOOL|UINT8|UINT16|UINT32|UINT64|INT8|INT16|INT32|INT64|FP32|FP64|FC32|FC64)$"
)


def uop_group(reg):
    srs = []
    for n in filter(None, [reg.match(i) for i in dir(lib)]):
        prefix, op, typ = n.groups()
        srs.append(UnaryOp(op, typ, getattr(lib, n.string)))
    return srs


def build_unaryops(__pdoc__):
    import tempfile

    this = sys.modules[__name__]
    for r in chain(uop_group(uop_re)):
        setattr(this, r.name, r)
        op, typ = r.name.split("_")
        f = tempfile.TemporaryFile()
        r.print(f=f)
        f.seek(0)
        __pdoc__[f"{typ}.{op}"] = f"""```{str(f.read(), 'utf8')}```"""


def _uop_name(name):  # pragma: nocover
    return "_{0}_uop_function".format(name)


def _build_uop_def(name, arg_type, result_type):  # pragma: nocover
    decl = dedent(
        """
    typedef void (*{0})({1}*, {1}*);
    """.format(
            _uop_name(name), arg_type, result_type
        )
    )
    return decl


def unary_op(arg_type):
    """Decorator to jit-compile Python function into a GrB_BinaryOp
    object.

    >>> from random import random
    >>> from pygraphblas import Matrix, binary_op, types, gviz
    >>> @unary_op(types.FP64)
    ... def random(x):
    ...     return x * random()
    >>> A = Matrix.dense(types.FP64, 3, 3, fill=1)
    >>> A.apply(random, out=A) is A
    True

    >>> ga = gviz.draw_matrix(A, scale=40,
    ...     filename='docs/imgs/unary_op_A')

    ![unary_op_A.png](../imgs/unary_op_A.png)

    """

    def inner(func):
        func_name = func.__name__
        sig = numba.void(
            numba.types.CPointer(arg_type._numba_t),
            numba.types.CPointer(arg_type._numba_t),
        )
        jitfunc = numba.jit(func, nopython=True)

        @numba.cfunc(sig, nopython=True)
        def wrapper(z, x):  # pragma: no cover
            result = jitfunc(x[0])
            z[0] = result

        out = core_ffi.new("GrB_UnaryOp*")
        lib.GrB_UnaryOp_new(
            out,
            core_ffi.cast("GxB_unary_function", wrapper.address),
            arg_type._gb_type,
            arg_type._gb_type,
        )

        return UnaryOp(func_name, arg_type.__name__, out[0])

    return inner



"""Pythonic GraphBLAS type wrappers including User Defined Types.

GraphBLAS has 13 built-in scalar types: Boolean, single and double
precision floating-point (real and complex), and 8, 16, 32, and 64-bit
signed and unsigned integers.  In addition, user-defined scalar types
can be created from nearly any C `typedef`, as long as the entire type
fits in a fixed-size contiguous block of memory (of arbitrary size).
All of these types can be used to create GraphBLAS sparse matrices,
vectors, or scalars.

"""
from .base import lib, _check, ffi as core_ffi
from textwrap import dedent
from operator import methodcaller, itemgetter
from functools import partial
import numba
import numpy
from numba import cfunc, jit, carray
from numba.core.typing import cffi_utils as cffi_support

from cffi import FFI

__pdoc__ = {}

__all__ = [
    "Type",
    "BOOL",
    "INT8",
    "INT16",
    "INT32",
    "INT64",
    "UINT8",
    "UINT16",
    "UINT32",
    "UINT64",
    "FP32",
    "FP64",
    "FC32",
    "FC64",
    "binop",
    "promote",
]


def _gb_type_to_type(gb_type):
    return MetaType._gb_type_map[gb_type]


class MetaType(type):

    _gb_type_map = {}
    _type_gb_map = {}
    _gb_name_type_map = {}
    _dtype_gb_map = {}

    def __new__(meta, type_name, bases, attrs):
        if attrs.get("base", False):
            cls = super().__new__(meta, type_name, bases, attrs)
            return cls
        if "members" in attrs:  # pragma: nocover
            m = attrs["members"]
            cls_ffi = FFI()
            cls_ffi.cdef(build_udt_def(type_name, m))
            t = core_ffi.new("GrB_Type*")
            _check(lib.GrB_Type_new(t, cls_ffi.sizeof(type_name)))
            cffi_support.map_type(cls_ffi.typeof(type_name), use_record_dtype=True)
            attrs["_ffi"] = cls_ffi
            attrs["_gb_type"] = t[0]
            attrs["_c_type"] = type_name
            attrs["member_def"] = list(map(methodcaller("split"), m))
            attrs["base_name"] = "UDT"
        else:
            attrs["_ffi"] = core_ffi
            _gb_type_name = type_name

        cls = super().__new__(meta, type_name, bases, attrs)
        meta._gb_type_map[cls._gb_type] = cls
        meta._type_gb_map[cls] = cls._gb_type
        meta._dtype_gb_map[cls._numpy_t] = cls
        meta._gb_name_type_map[type_name] = cls
        meta._gb_name_type_map[cls._c_type] = cls

        cls._ptr = cls._c_type + "*"
        cls.default_zero = getattr(cls, "default_zero", core_ffi.NULL)
        cls.default_one = getattr(cls, "default_one", core_ffi.NULL)
        get = partial(getattr, lib)
        cls._base_name = base_name = getattr(cls, "_base_name", cls.__name__)
        cls._prefix = prefix = getattr(cls, "_prefix", "GrB")
        cls._Monoid_new = get("{}_Monoid_new_{}".format(prefix, base_name))
        cls._Matrix_setElement = get(
            "{}_Matrix_setElement_{}".format(prefix, base_name)
        )
        cls._Matrix_extractElement = get(
            "{}_Matrix_extractElement_{}".format(prefix, base_name)
        )
        cls._Matrix_extractTuples = get(
            "{}_Matrix_extractTuples_{}".format(prefix, base_name)
        )
        cls._Matrix_assignScalar = get("{}_Matrix_assign_{}".format(prefix, base_name))
        cls._Matrix_apply_BinaryOp1st = get(
            "{}_Matrix_apply_BinaryOp1st_{}".format(prefix, base_name)
        )
        cls._Matrix_apply_BinaryOp2nd = get(
            "{}_Matrix_apply_BinaryOp2nd_{}".format(prefix, base_name)
        )
        cls._Vector_setElement = get(
            "{}_Vector_setElement_{}".format(prefix, base_name)
        )
        cls._Vector_extractElement = get(
            "{}_Vector_extractElement_{}".format(prefix, base_name)
        )
        cls._Vector_extractTuples = get(
            "{}_Vector_extractTuples_{}".format(prefix, base_name)
        )
        cls._Vector_assignScalar = get("{}_Vector_assign_{}".format(prefix, base_name))
        cls._Vector_apply_BinaryOp1st = get(
            "{}_Vector_apply_BinaryOp1st_{}".format(prefix, base_name)
        )
        cls._Vector_apply_BinaryOp2nd = get(
            "{}_Vector_apply_BinaryOp2nd_{}".format(prefix, base_name)
        )
        cls._Scalar_setElement = get("GxB_Scalar_setElement_{}".format(base_name))
        cls._Scalar_extractElement = get(
            "GxB_Scalar_extractElement_{}".format(base_name)
        )
        return cls

    def new_monoid(cls, op, identity):
        from .monoid import Monoid

        monoid = core_ffi.new("GrB_Monoid[1]")
        if cls._base_name == "UDT":  # pragma: nocover
            i = cls._ffi.new(cls._ptr)
            i[0] = identity
            identity = i
        _check(cls._Monoid_new(monoid, op.binaryop, identity))
        return Monoid("PLUS", cls.__name__, monoid[0], udt=cls)

    def new_semiring(cls, monoid, op):
        from .semiring import Semiring

        semiring = core_ffi.new("GrB_Semiring[1]")
        _check(
            lib.GrB_Semiring_new(
                semiring, monoid.get_monoid(), op.get_binaryop(core_ffi.NULL)
            )
        )
        return Semiring("PLUS", "TIMES", cls.__name__, semiring[0], udt=cls)

    def gb_from_name(cls, name):
        return cls._gb_name_type_map[name]._gb_type


class Type(metaclass=MetaType):
    default_one = 1
    """The default value used to represent 1 for filling in types."""
    default_zero = 0
    """The default value used to represent 0 for filling in types."""
    base = True
    _typecode = None

    @classmethod
    def format_value(cls, val, width=2, prec=None):
        """Return the value as a formatted string for display."""
        return f"{val:{width}}"

    @classmethod
    def _default_addop(cls):
        return cls.PLUS

    @classmethod
    def _default_multop(cls):
        return cls.TIMES

    @classmethod
    def _default_semiring(cls):
        return cls.PLUS_TIMES

    @classmethod
    def _from_value(cls, value):
        """"""
        if cls._base_name != "UDT":
            return value
        else:  # pragma: nocover
            data = cls._ffi.new("%s[1]" % cls.__name__)
            for (_, name), val in zip(cls.member_def, value):
                setattr(data[0], name, val)
            return data

    @classmethod
    def _to_value(cls, cdata):
        if cls._base_name != "UDT":
            return cdata
        else:  # pragma: nocover
            return tuple(getattr(cdata, name) for (_, name) in cls.member_def)


class BOOL(Type):
    """GraphBLAS Boolean Type."""

    _gb_type = lib.GrB_BOOL
    _c_type = "_Bool"
    default_one = True
    default_zero = False
    _typecode = "B"
    _numba_t = numba.boolean
    _numpy_t = numpy.bool_

    @classmethod
    def _default_addop(self):  # pragma: nocover
        return self.LOR

    @classmethod
    def _default_multop(self):  # pragma: nocover
        return self.LAND

    @classmethod
    def _default_semiring(self):  # pragma: nocover
        return self.LOR_LAND

    @classmethod
    def format_value(cls, val, width=2, prec=None):
        f = "{:>%s}" % width
        if not isinstance(val, bool):
            return f.format(val)
        return f.format("t") if val is True else f.format("f")

    @classmethod
    def _to_value(cls, cdata):
        return bool(cdata)


class INT8(Type):
    """GraphBLAS 8 bit signed integer."""

    _gb_type = lib.GrB_INT8
    _c_type = "int8_t"
    _typecode = "b"
    _numba_t = numba.int8
    _numpy_t = numpy.int8


class UINT8(Type):
    """GraphBLAS 8 bit unsigned integer."""

    _gb_type = lib.GrB_UINT8
    _c_type = "uint8_t"
    _typecode = "B"
    _numba_t = numba.uint8
    _numpy_t = numpy.uint8


class INT16(Type):
    """GraphBLAS 16 bit signed integer."""

    _gb_type = lib.GrB_INT16
    _c_type = "int16_t"
    _typecode = "i"
    _numba_t = numba.int16
    _numpy_t = numpy.int16


class UINT16(Type):
    """GraphBLAS 16 bit unsigned integer."""

    _gb_type = lib.GrB_UINT16
    _c_type = "uint16_t"
    _typecode = "I"
    _numba_t = numba.uint16
    _numpy_t = numpy.uint16


class INT32(Type):
    """GraphBLAS 32 bit signed integer."""

    _gb_type = lib.GrB_INT32
    _c_type = "int32_t"
    _typecode = "l"
    _numba_t = numba.int32
    _numpy_t = numpy.int32


class UINT32(Type):
    """GraphBLAS 32 bit unsigned integer."""

    _gb_type = lib.GrB_UINT32
    _c_type = "uint32_t"
    _typecode = "L"
    _numba_t = numba.uint32
    _numpy_t = numpy.uint32


class INT64(Type):
    """GraphBLAS 64 bit signed integer."""

    _gb_type = lib.GrB_INT64
    _c_type = "int64_t"
    _typecode = "q"
    _numba_t = numba.int64
    _numpy_t = numpy.int64


class UINT64(Type):
    """GraphBLAS 64 bit unsigned integer."""

    _gb_type = lib.GrB_UINT64
    _c_type = "uint64_t"
    _typecode = "Q"
    _numba_t = numba.uint64
    _numpy_t = numpy.uint64


class FP32(Type):
    """GraphBLAS 32 bit float."""

    default_one = 1.0
    default_zero = 0.0
    _gb_type = lib.GrB_FP32
    _c_type = "float"
    _typecode = "f"
    _numba_t = numba.float32
    _numpy_t = numpy.float32

    @classmethod
    def format_value(cls, val, width=2, prec=2):
        return f"{val:>{width}.{prec}}"


class FP64(Type):
    """GraphBLAS 64 bit float."""

    default_one = 1.0
    default_zero = 0.0
    _gb_type = lib.GrB_FP64
    _c_type = "double"
    _typecode = "d"
    _numba_t = numba.float64
    _numpy_t = numpy.float64

    @classmethod
    def format_value(cls, val, width=2, prec=2):
        return f"{val:>{width}.{prec}}"


class FC32(Type):
    """GraphBLAS 32 bit float complex."""

    _prefix = "GxB"
    default_one = complex(1.0)
    default_zero = complex(0.0)
    _gb_type = lib.GxB_FC32
    _c_type = "float _Complex"
    _numba_t = numba.complex64
    _numpy_t = numpy.complex64


class FC64(Type):
    """GraphBLAS 64 bit float complex."""

    _prefix = "GxB"
    default_one = complex(1.0)
    default_zero = complex(0.0)
    _gb_type = lib.GxB_FC64
    _c_type = "double _Complex"
    _numba_t = numba.complex128
    _numpy_t = numpy.complex128


def _gb_from_type(typ):  # pragma: nocover
    if typ is int:
        return INT64
    if typ is float:
        return FP64
    if typ is bool:
        return BOOL
    if typ is complex:  # pragma: no cover
        return FC64
    raise TypeError(f"cannot turn {typ!r} into GraphBLAS type.")


def udt_head(name):  # pragma: nocover
    return dedent(
        """
    typedef struct %s {
    """
        % name
    )


def udt_body(members):  # pragma: nocover
    return ";\n".join(members) + ";"


def udt_tail(name):  # pragma: nocover
    return dedent(
        """
    } %s;
    """
        % name
    )


def build_udt_def(typ, members):  # pragma: nocover
    return udt_head(typ) + udt_body(members) + udt_tail(typ)


def binop_name(typ, name):  # pragma: nocover
    return "{0}_{1}_binop_function".format(typ, name)


def build_binop_def(typ, name, boolean=False):  # pragma: nocover
    if boolean:
        return dedent(
            """
        typedef void (*{0})(bool*, {1}*, {1}*);
        """.format(
                binop_name(typ, name), typ
            )
        )
    return dedent(
        """
    typedef void (*{0})({1}*, {1}*, {1}*);
    """.format(
            binop_name(typ, name), typ
        )
    )


def binop(boolean=False):  # pragma: nocover
    from .binaryop import BinaryOp

    class inner:
        def __init__(self, func):
            self.func = func

        def __set_name__(self, cls, name):
            func_name = self.func.__name__
            cls_name = cls.__name__
            if cls._base_name == "UDT":
                cls._ffi.cdef(build_binop_def(cls_name, func_name, boolean))
                sig = cffi_support.map_type(
                    cls._ffi.typeof(binop_name(cls_name, func_name)),
                    use_record_dtype=True,
                )
            else:
                sig = numba.void(cls._numba_t, cls._numba_t, cls._numba_t)

            jitfunc = jit(self.func, nopython=True)

            @cfunc(sig)
            def wrapper(z_, x_, y_):
                z = carray(z_, 1)[0]
                x = carray(x_, 1)[0]
                y = carray(y_, 1)[0]
                jitfunc(z, x, y)

            self.op = BinaryOp(func_name, cls_name, wrapper, cls, boolean)
            setattr(cls, func_name, self.op)

    return inner


def get_add(sring):
    monoid = core_ffi.new("GrB_Monoid*")
    _check(lib.GxB_Semiring_add(monoid, sring))
    return monoid[0]


def get_binaryop(moid):
    op = core_ffi.new("GrB_BinaryOp*")
    _check(lib.GxB_Monoid_operator(op, moid))
    return op[0]


def get_ztype(bop):
    typ = core_ffi.new("GrB_Type*")
    _check(lib.GxB_BinaryOp_ztype(typ, bop))
    return _gb_type_to_type(typ[0])


def get_semiring_ztype(sring):
    return get_ztype(get_binaryop(get_add(sring)))


_int_types = (INT8, INT16, INT32, INT64, UINT8, UINT16, UINT32, UINT64)

_float_types = (FP32, FP64)

_promotion_order = (
    FC64,
    FC32,
    FP64,
    FP32,
    INT64,
    UINT64,
    INT32,
    UINT32,
    INT16,
    UINT16,
    INT8,
    UINT8,
)


def promote(left, right):
    """Do type promotion, determine the result type of an operation
    infered from the operands and possible semiring.

    """
    if left == right:
        return left
    elif left == BOOL:
        return right
    elif right == BOOL:
        return left
    for t in _promotion_order:
        if left == t or right == t:
            return t
    raise TypeError(
        "inconvertable types %s and %s" % (repr(left), repr(right))
    )  # pragma: nocover



"""# Installation

pygraphblas requires the
[SuiteSparse:GraphBLAS](http://faculty.cse.tamu.edu/davis/GraphBLAS.html)
library.  The easiest way currently to install pygraphblas on Linux is
to use pip:

    pip install pygraphblas

This is currently only supported on Linux but Windows and MacOS binary
packages are currently a work in progress.

To install from source, first install SuiteSparse, then run:

    python setup.py install
  
There are two ways to download precompiled binaries of pygraphblas
with SuiteSparse included.  One way is to use `pip install
pygraphblas` on an Intel Linux machine.  This will download a package
compatible with most modern linux distributions.  This also works in a
Docker container on Mac.

There are also pre-build docker images based on Ubuntu 20.04 that have
a pre-compiled SuiteSparse and pygraphblas installed.  These come in
two flavors `minimal` which is the Ipython interpreter-only, and
`notebook` which comes with a complete Jupyter Notebook server.  These
containers also work on Mac.

An installation script for Ubuntu 18.04 is provided in the
`install-ubuntu.sh` file.

NOTE: DO NOT USE THESE PRE-COMPILED BINARIES FOR BENCHMARKING
SUITESPARSE.  These binaries are not guaranteed to be idealy compiled
for your environment.

## Docker 

pygraphblas is distributed as two different docker images on [Docker
Hub](https://cloud.docker.com/repository/docker/pygraphblas/pygraphblas/general)
.  The "minimal" image, containing only the library and
[ipython](https://ipython.org/) and can be run with the command:

    docker run --rm -it graphblas/pygraphblas-minimal ipython

You can run a "full" [Jupyter notebook](https://jupyter.org/) server
with docker and try the example Notebooks use the command:

    docker run --rm -it graphblas/pygraphblas-notebook

Open up the URL printed on your terminal screen, typically something
liker `http://127.0.0.1:8888/tree/pygraphblas/demo` and see the
following Notebooks:

 - [Introduction to GraphBLAS with Python](https://github.com/Graphegon/pygraphblas/blob/main/demo/Introduction-to-GraphBLAS-with-Python.ipynb)
 - [PageRank](https://github.com/Graphegon/pygraphblas/blob/main/demo/PageRank.ipynb)
 - [Betweeness Centrality](https://github.com/Graphegon/pygraphblas/blob/main/demo/BetweenessCentrality.ipynb)
 - [K-Truss Subgraphs](https://github.com/Graphegon/pygraphblas/blob/main/demo/K-Truss.ipynb)
 - [Triangle Counting](https://github.com/Graphegon/pygraphblas/blob/main/demo/Triangle-Counting.ipynb)
 - [Louvain Community Detection](https://github.com/Graphegon/pygraphblas/blob/main/demo/Louvain.ipynb)
 - [RadiX-Net Topologies](https://github.com/Graphegon/pygraphblas/blob/main/demo/RadiX-Net-with-pygraphblas.ipynb)
 - [User Defined Types](https://github.com/Graphegon/pygraphblas/blob/main/demo/User-Defined-Types.ipynb)
 - [Log Semiring Type](https://github.com/Graphegon/pygraphblas/blob/main/demo/Log-Semiring.ipynb)
 - [Kronecker Power Graphs](https://github.com/Graphegon/pygraphblas/blob/main/demo/Sierpinski-Graph.ipynb)

# Tests

To run the tests checkout pygraphblas and use:

    $ ./test.sh

# Summary

pygraphblas is a python extension that bridges [The GraphBLAS
API](http://graphblas.org) with the [Python](https://python.org)
programming language.  It uses the
[CFFI](https://cffi.readthedocs.io/en/latest/) library to wrap the low
level GraphBLAS API and provides high level Matrix and Vector Python
types that make GraphBLAS simple and easy.

GraphBLAS is a sparse linear algebra API optimized for processing
graphs encoded as sparse matrices and vectors.  In addition to common
real/integer matrix algebra operations, GraphBLAS supports over a
thousand different [Semiring](https://en.wikipedia.org/wiki/Semiring)
algebra operations, that can be used as basic building blocks to
implement a wide variety of graph algorithms. See
[Applications](https://en.wikipedia.org/wiki/Semiring#Applications)
from Wikipedia for some specific examples.

pygraphblas leverages the expertise in the field of sparse matrix
programming by [The GraphBLAS Forum](http://graphblas.org) and uses
the
[SuiteSparse:GraphBLAS](http://faculty.cse.tamu.edu/davis/GraphBLAS.html)
API implementation. SuiteSparse:GraphBLAS is brought to us by the work
of [Dr. Tim Davis](http://faculty.cse.tamu.edu/davis/welcome.html),
professor in the Department of Computer Science and Engineering at
Texas A&M University.  [News and
information](http://faculty.cse.tamu.edu/davis/news.html) can provide
you with a lot more background information, in addition to the
references below.

While it is my goal to make it so that pygraphblas works with any
GraphBLAS implementation, it currently only works with SuiteSparse.
SuiteSparse is currently the only realistically usable GraphBLAS
implementation, and additionally it provides several "extension"
features and pre-packaged objects that are very useful for
pygraphblas.  If there is a GraphBLAS implementation you would like to
see support for in pygraphblas, please consider creating an issue for
it for discussion and/or sending me a pull request.

# Introduction to Graphs and Matrices

GraphBLAS uses matrices and Linear Algebra to represent graphs, as
described [in this mathmatical introduction to
GraphBLAS](http://www.mit.edu/~kepner/GraphBLAS/GraphBLAS-Math-release.pdf)
by [Dr. Jeremy Kepner](http://www.mit.edu/~kepner/) head and founder
of [MIT Lincoln Laboratory Supercomputing
Center](http://news.mit.edu/2016/lincoln-laboratory-establishes-supercomputing-center-0511).

There are two useful matrix representations of graphs: [Adjacency
Matrices](https://en.wikipedia.org/wiki/Adjacency_matrix) and
[Incidence Matrices](https://en.wikipedia.org/wiki/Incidence_matrix).
For this introduction we will focus on the adjacency type as they are
simpler, but the same ideas apply to both, both are suported by
GraphBLAS and pygraphblas, and it is easy to switch back and forth
between them.

![An example graph and adjacency matrix](../imgs/AdjacencyMatrixBFS.png)

On the left is a graph, and on the right, the adjacency matrix that
represents it. The matrix has a row and column for every node in the
graph.  If there is an edge going from node A to B, then there will be
a value present in the intersection of As row with Bs column.  How it
differs from many other matrix representations is that the matrix is
sparse, nothing is stored in computer memory where there are unused
elements.

Sparsity is important because one practical problem with
matrix-encoding graphs is that most real-world graphs tend to be
sparse, as above, only 7 of 36 possible elements have a value. Those
that have values tend to be scattered randomly across the matrix
(for "typical" graphs), so dense linear algebra libraries like BLAS or
numpy do not encode or operate on them efficiently, as the relevant
data is mostly empty memory with actual data elements spaced far
apart.  This wastes memory and CPU resources, and defeats CPU caching
mechanisms.

For example, suppose a fictional social network has 1 billion users,
and each user has about 100 friends, which means there are about 100
billion (1e+11) connections in the graph.  A dense matrix large enough
to hold this graph would need (1 billion)^2 or
(1,000,000,000,000,000,000), a "quintillion" elements, but only 1e+11
of them would have meaningful values, leaving only 0.0000001th of the
matrix being utilized.

By using a sparse matrix instead of dense, only the elements used are
actually stored in memory. The parts of the matrix with no value are
*interpreted*, but not necessarily stored, as an identity value, which
may or may not be the actual number zero, but possibly other values
like positive or negative infinity depending on the particular
semiring operations applied to the matrix.

Semirings encapsulate different algebraic operations and identities
that can be used to multiply matrices and vectors.  Anyone who has
multiplied matrices has used at least one Semiring before, typically
referred to as "plus_times".  This is the common operation of
multiplying two matrices containing real numbers, the corresponding row
and column entries are multipled and the results are summed for the
final value.

# Using the API

The core idea of the GraphBLAS is the mathematical duality between a
graph and a `pygraphblas.Matrix`.  There are a few ways to contstruct
matricies, but a simple approach is to provide three lists of data,
the first are are lists of the row and column positions that define
the begining and end of a graph edge, and the third list is the weight
for that edge:

>>> import pygraphblas as gb
>>> I = [0, 0, 1, 1, 2, 3, 3, 4, 5, 6, 6, 6]
>>> J = [1, 3, 4, 6, 5, 0, 2, 5, 2, 2, 3, 4]
>>> V = [True] * len(I)
>>> M = gb.Matrix.from_lists(I, J, V)
>>> print(M)
      0  1  2  3  4  5  6
  0|     t     t         |  0
  1|              t     t|  1
  2|                 t   |  2
  3|  t     t            |  3
  4|                 t   |  4
  5|        t            |  5
  6|        t  t  t      |  6
      0  1  2  3  4  5  6

>>> from pygraphblas import gviz
>>> g = gviz.draw_graph(M, show_weight=False, 
...     filename='docs/imgs/Matrix_from_lists2')

![Matrix_from_lists2.png](../imgs/Matrix_from_lists2.png)

>>> g = gviz.draw_matrix(M, scale=40, 
...     filename='docs/imgs/Matrix_from_lists_matrix')

![Matrix_from_lists_matrix.png](../imgs/Matrix_from_lists_matrix.png)

GraphBLAS is a sparse [Linear
Algebra](https://en.wikipedia.org/wiki/Linear_algebra) API optimized
for processing graphs encoded as sparse matrices and vectors.  In
addition to common real/integer `pygraphblas.Matrix` algebra
operations, GraphBLAS supports over a thousand different
[Semiring](https://en.wikipedia.org/wiki/Semiring) algebra operations,
that can be used as basic building blocks to implement a wide variety
of graph algorithms. See
[Applications](https://en.wikipedia.org/wiki/Semiring#Applications)
from Wikipedia for some specific examples.

The core operation of Linear Algebra is [Matrix
Multiplication](https://en.wikipedia.org/wiki/Matrix_multiplication).
In this GraphBLAS duality, this is an operation along the edges of a
graph from nodes to their adjacenct neighbors, taking one step in a
[Breadth First
Search](https://en.wikipedia.org/wiki/Breadth-first_search) across the
graph:

>>> v = gb.Vector.from_lists([0], [True], M.nrows)
>>> y = v @ M
>>> g = gviz.draw_matrix_op(v, '@', M, y, scale=40, labels=True, 
... column=False, filename='docs/imgs/bfs_step')

![bfs_step.png](../imgs/bfs_step.png)

pygraphblas leverages the expertise in the field of sparse matrix
programming by [The GraphBLAS Forum](http://graphblas.org) and uses
the
[SuiteSparse:GraphBLAS](http://faculty.cse.tamu.edu/davis/GraphBLAS.html)
API implementation. SuiteSparse:GraphBLAS is brought to us by the work
of [Dr. Tim Davis](http://faculty.cse.tamu.edu/davis/welcome.html),
professor in the Department of Computer Science and Engineering at
Texas A&M University.  [News and
information](http://faculty.cse.tamu.edu/davis/news.html) can provide
you with a lot more background information.

"""

from suitesparse_graphblas import lib, ffi, initialize, is_initialized


def init(blocking=False):
    initialize(blocking=blocking, memory_manager="c")


if not is_initialized():  # pragma: nocover
    init()

from .base import (
    lib,
    ffi,
    GxB_INDEX_MAX,
    GxB_IMPLEMENTATION,
    GxB_SPEC,
    options_get,
    options_set,
)


IMPLEMENTATION_MAJOR = lib.GxB_IMPLEMENTATION_MAJOR
IMPLEMENTATION_MINOR = lib.GxB_IMPLEMENTATION_MINOR
IMPLEMENTATION_SUB = lib.GxB_IMPLEMENTATION_SUB
IMPLEMENTATION_VERSION = (
    IMPLEMENTATION_MAJOR,
    IMPLEMENTATION_MINOR,
    IMPLEMENTATION_SUB,
)

from . import _version

__version__ = _version.get_versions()["version"]


def get_version():  # pragma: nocover
    """Return the pygraphblas version."""
    return __version__


from .semiring import build_semirings
from .binaryop import build_binaryops, Accum, binary_op
from .unaryop import build_unaryops, unary_op
from .selectop import build_selectops, select_op
from .monoid import build_monoids
from .matrix import Matrix
from .vector import Vector
from .scalar import Scalar
from . import descriptor

__pdoc__ = {
    "base": False,
    "build": False,
    "unaryop": False,
    "binaryop": False,
    "monoid": False,
    "semiring": False,
    "matrix": False,
    "vector": False,
    "scalar": False,
    "types": False,
    "run_doctests": False,
    "descriptor": True,
    "selectop": True,
}

build_semirings(__pdoc__)
build_binaryops(__pdoc__)
build_unaryops(__pdoc__)
build_monoids(__pdoc__)
build_selectops(__pdoc__)

from .types import (
    BOOL,
    FP64,
    FP32,
    FC64,
    FC32,
    INT64,
    INT32,
    INT16,
    INT8,
    UINT64,
    UINT32,
    UINT16,
    UINT8,
)

__all__ = [
    "GxB_INDEX_MAX",
    "GxB_IMPLEMENTATION",
    "GxB_SPEC",
    "Matrix",
    "Vector",
    "Scalar",
    "Accum",
    "BOOL",
    "FP64",
    "FP32",
    "FC64",
    "FC32",
    "INT64",
    "INT32",
    "INT16",
    "INT8",
    "UINT64",
    "UINT32",
    "UINT16",
    "UINT8",
    "descriptor",
    "selectop",
    "binary_op",
    "unary_op",
    "select_op",
    "options_set",
    "options_get",
]

GxB_INDEX_MAX = GxB_INDEX_MAX
"""Maximum key size for SuiteSparse, defaults to `2**60`."""

GxB_IMPLEMENTATION = GxB_IMPLEMENTATION
""" Tuple containing GxB_IMPLEMENTATION (MAJOR, MINOR, SUB) """

GxB_SPEC = GxB_SPEC
""" Tuple containing GxB_SPEC (MAJOR, MINOR, SUB) """


def run_doctests(raise_on_error=False):
    from . import matrix
    from . import vector
    from . import descriptor
    from . import base
    from . import unaryop
    from . import binaryop
    import sys, doctest

    this = sys.modules[__name__]
    for mod in (
        this,
        selectop,
        unaryop,
        binaryop,
        matrix,
        vector,
        descriptor,
        base,
    ):
        doctest.testmod(
            mod, optionflags=doctest.ELLIPSIS, raise_on_error=raise_on_error
        )


from . import _version

__version__ = _version.get_versions()["version"]



"""Contains all automatically generated BinaryOps from CFFI.

"""

__all__ = [
    "BinaryOp",
    "Accum",
    "current_binop",
    "current_accum",
    "binary_op",
]

import sys
import re
import contextvars
from itertools import chain
from collections import defaultdict
from functools import partial
import numba

from .base import lib, ffi, _check
from . import types

current_accum = contextvars.ContextVar("current_accum")
current_binop = contextvars.ContextVar("current_binop")


class BinaryOp:
    """Wrapper around GrB_BinaryOp."""

    def __init__(self, op, typ, binaryop, udt=None, boolean=False):
        if udt is not None:  # pragma: no cover
            o = ffi.new("GrB_BinaryOp*")
            udt = udt._gb_type
            lib.GrB_BinaryOp_new(
                o,
                ffi.cast("GxB_binary_function", binaryop.address),
                lib.GrB_BOOL if boolean else udt,
                udt,
                udt,
            )
            self.binaryop = o[0]
        else:
            self.binaryop = binaryop
            cls = getattr(types, typ)
            setattr(cls, op, self)
            setattr(cls, op.lower(), self)
        self.name = "_".join((op, typ))
        self.__doc__ = self.name
        self.token = None

    def __enter__(self):
        self.token = current_binop.set(self)
        return self

    def __exit__(self, *errors):  # pragma: nocover
        current_binop.reset(self.token)
        return False

    def get_binaryop(self, left=None, right=None):
        return self.binaryop

    def print(self, level=2, name="", f=sys.stdout):  # pragma: nocover
        """Print the matrix using `GxB_Matrix_fprint()`, by default to
        `sys.stdout`.

        Level 1: Short description
        Level 2: Short list, short numbers
        Level 3: Long list, short number
        Level 4: Short list, long numbers
        Level 5: Long list, long numbers

        """
        _check(lib.GxB_BinaryOp_fprint(self.binaryop, bytes(name, "utf8"), level, f))


class Accum:
    """Helper context manager to specify accumulator binary operator in
    overloaded operator contexts like `@`.  This disambiguates for
    methods like `Matrix.eadd` and `Matrix.emult` that can specify
    both a binary operators *and* a binary accumulator.

    See those methods and `Matrix.mxm` for examples.

    """

    __slots__ = ("binaryop", "token")

    def __init__(self, binaryop):
        self.binaryop = binaryop

    def __enter__(self):
        self.token = current_accum.set(self.binaryop)
        return self

    def __exit__(self, *errors):
        current_accum.reset(self.token)
        return False


grb_binop_re = re.compile(
    "^(GrB|GxB)_(FIRST|SECOND|MIN|MAX|PLUS|MINUS|RMINUS|TIMES|DIV|RDIV|"
    "FIRSTI|FIRSTI1|FIRSTJ|FIRSTJ1|SECONDI|SECONDI1|SECONDJ|SECONDJ1|"
    "PAIR|ANY|POW|EQ|NE|GT|LT|GE|LE|LOR|LAND|LXOR|BOR|BAND|BXOR|BXNOR|"
    "ATAN2|HYPOT|FMOD|REMAINDER|LDEXP|COPYSIGN|BGET|BSET|BCLR|BSHIFT|CMPLX)_"
    "(BOOL|UINT8|UINT16|UINT32|UINT64|INT8|INT16|INT32|INT64|FP32|FP64|FC32|FC64)$"
)

pure_bool_re = re.compile("^(GrB|GxB)_(LOR|LAND|LXOR)_(BOOL)$")


def binop_group(reg):
    srs = []
    for n in filter(None, [reg.match(i) for i in dir(lib)]):
        prefix, op, typ = n.groups()
        srs.append(BinaryOp(op, typ, getattr(lib, n.string)))
    return srs


def build_binaryops(__pdoc__):
    import tempfile

    this = sys.modules[__name__]
    for r in chain(binop_group(grb_binop_re), binop_group(pure_bool_re)):
        setattr(this, r.name, r)
        this.__all__.append(r.name)
        op, typ = r.name.split("_")
        f = tempfile.TemporaryFile()
        r.print(f=f)
        f.seek(0)
        __pdoc__[f"{typ}.{op}"] = f"""```{str(f.read(), 'utf8')}```"""


def binary_op(arg_type, nopython=True):
    """Decorator to jit-compile Python function into a GrB_BinaryOp
    object.

    >>> from random import uniform
    >>> from pygraphblas import Matrix, binary_op, types, gviz
    >>> @binary_op(types.FP64)
    ... def uniform(x, y):
    ...     return uniform(x, y)
    >>> A = Matrix.dense(types.FP64, 3, 3, fill=0)
    >>> B = A.dup()
    >>> with uniform:
    ...     A += 1

    Calling `A += 1` with the `uniform` binary operator is the same as
    calling `apply_second` with an `out` parameter:

    >>> B.apply_second(uniform, 1, out=B) is B
    True
    >>> ga = gviz.draw_matrix(A, scale=40,
    ...     filename='docs/imgs/binary_op_A')
    >>> gb = gviz.draw_matrix(B, scale=40,
    ...     filename='docs/imgs/binary_op_B')


    ![binary_op_A.png](../imgs/binary_op_A.png) ![binary_op_B.png](../imgs/binary_op_B.png)

    """

    def inner(func):
        func_name = func.__name__
        sig = numba.void(
            numba.types.CPointer(arg_type._numba_t),
            numba.types.CPointer(arg_type._numba_t),
            numba.types.CPointer(arg_type._numba_t),
        )
        jitfunc = numba.jit(func, nopython=nopython)

        @numba.cfunc(sig, nopython=True)
        def wrapper(z, x, y):  # pragma: no cover
            result = jitfunc(x[0], y[0])
            z[0] = result

        out = ffi.new("GrB_BinaryOp*")
        lib.GrB_BinaryOp_new(
            out,
            ffi.cast("GxB_binary_function", wrapper.address),
            arg_type._gb_type,
            arg_type._gb_type,
            arg_type._gb_type,
        )

        return BinaryOp(func_name, arg_type.__name__, out[0])

    return inner
